from o3seespy.command.uniaxial_material.base_material import UniaxialMaterialBase



class Concrete01(UniaxialMaterialBase):
    """
    The Concrete01 UniaxialMaterial Class
    
    This command is used to construct a uniaxial Kent-Scott-Park concrete material object with degraded linear
    unloading/reloading stiffness according to the work of Karsan-Jirsa and no tensile strength. (REF: Fedeas).
    """
    op_type = 'Concrete01'

    def __init__(self, osi, fpc, epsc0, fpcu, eps_u):
        """
        Initial method for Concrete01

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fpc: float
            Concrete compressive strength at 28 days (compression is negative)
        epsc0: float
            Concrete strain at maximum strength
        fpcu: float
            Concrete crushing strength
        eps_u: float
            Concrete strain at crushing strength

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.Concrete01(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0)
        """
        self.osi = osi
        self.fpc = float(fpc)
        self.epsc0 = float(epsc0)
        self.fpcu = float(fpcu)
        self.eps_u = float(eps_u)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fpc, self.epsc0, self.fpcu, self.eps_u]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_fc(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'fc', value, ele, eles)

    def set_epsco(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'epsco', value, ele, eles)

    def set_fcu(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'fcu', value, ele, eles)

    def set_epscu(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'epscu', value, ele, eles)


class Concrete02(UniaxialMaterialBase):
    """
    The Concrete02 UniaxialMaterial Class
    
    This command is used to construct a uniaxial Kent-Scott-Park concrete material object with degraded linear
    unloading/reloading stiffness according to the work of Karsan-Jirsa and no tensile strength. (REF: Fedeas).
    """
    op_type = 'Concrete02'

    def __init__(self, osi, fpc, epsc0, fpcu, eps_u, lamb, ft, ets):
        """
        Initial method for Concrete02

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fpc: float
            Concrete compressive strength at 28 days (compression is negative)
        epsc0: float
            Concrete strain at maximum strength
        fpcu: float
            Concrete crushing strength
        eps_u: float
            Concrete strain at crushing strength
        lamb: float
            Ratio between unloading slope at $epscu and initial slope
        ft: float
            Tensile strength
        ets: float
            Tension softening stiffness (absolute value) (slope of the linear tension softening branch)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.Concrete02(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0, lamb=1.0, ft=1.0, ets=1.0)
        """
        self.osi = osi
        self.fpc = float(fpc)
        self.epsc0 = float(epsc0)
        self.fpcu = float(fpcu)
        self.eps_u = float(eps_u)
        self.lamb = float(lamb)
        self.ft = float(ft)
        self.ets = float(ets)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fpc, self.epsc0, self.fpcu, self.eps_u, self.lamb, self.ft, self.ets]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class Concrete04(UniaxialMaterialBase):
    """
    The Concrete04 UniaxialMaterial Class
    
    This command is used to construct a uniaxial Popovics concrete material object with degraded linear
    unloading/reloading stiffness according to the work of Karsan-Jirsa and tensile strength with exponential decay.
    """
    op_type = 'Concrete04'

    def __init__(self, osi, fc, epsc, epscu, ec, fct, et, beta):
        """
        Initial method for Concrete04

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc: float
            Floating point values defining concrete compressive strength at 28 days (compression is negative)
        epsc: float
            Floating point values defining concrete strain at maximum strength
        epscu: float
            Floating point values defining concrete strain at crushing strength
        ec: float
            Floating point values defining initial stiffness
        fct: float
            Floating point value defining the maximum tensile strength of concrete 
        et: float
            Floating point value defining ultimate tensile strain of concrete 
        beta: float
            Loating point value defining the exponential curve parameter to define the residual stress (as a factor of
            ft) at etu

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.Concrete04(osi, fc=1.0, epsc=1.0, epscu=1.0, ec=1.0, fct=1.0, et=1.0, beta=1.0)
        """
        self.osi = osi
        self.fc = float(fc)
        self.epsc = float(epsc)
        self.epscu = float(epscu)
        self.ec = float(ec)
        self.fct = float(fct)
        self.et = float(et)
        self.beta = float(beta)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc, self.epsc, self.epscu, self.ec, self.fct, self.et, self.beta]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class Concrete06(UniaxialMaterialBase):
    """
    The Concrete06 UniaxialMaterial Class
    
    This command is used to construct a uniaxial concrete material object with tensile strength, nonlinear tension
    stiffening and compressive behavior based on Thorenfeldt curve.
    """
    op_type = 'Concrete06'

    def __init__(self, osi, fc, e0, n, k, alpha1, fcr, ecr, b, alpha2):
        r"""
        Initial method for Concrete06

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc: float
            Concrete compressive strength (compression is negative)
        e0: float
            Strain  at compressive strength
        n: float
            Compressive shape factor
        k: float
            Post-peak compressive shape factor
        alpha1: float
            :math:`\alpha_1` parameter for compressive plastic strain definition
        fcr: float
            Tensile strength
        ecr: float
            Tensile strain at peak stress (fcr)
        b: float
            Exponent of the tension stiffening curve
        alpha2: float
            :math:`\alpha_2` parameter for tensile plastic strain definition

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.Concrete06(osi, fc=1.0, e0=1.0, n=1.0, k=1.0, alpha1=1.0, fcr=1.0, ecr=1.0, b=1.0, alpha2=1.0)
        """
        self.osi = osi
        self.fc = float(fc)
        self.e0 = float(e0)
        self.n = float(n)
        self.k = float(k)
        self.alpha1 = float(alpha1)
        self.fcr = float(fcr)
        self.ecr = float(ecr)
        self.b = float(b)
        self.alpha2 = float(alpha2)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc, self.e0, self.n, self.k, self.alpha1, self.fcr, self.ecr, self.b, self.alpha2]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class Concrete07(UniaxialMaterialBase):
    """
    The Concrete07 UniaxialMaterial Class
    
    Concrete07 is an implementation of Chang & Mander's 1994 concrete model with simplified unloading and reloading
    curves. Additionally the tension envelope shift with respect to the origin proposed by Chang and Mander has been
    removed. The model requires eight input parameters to define the monotonic envelope of confined and unconfined
    concrete in the following form:
    """
    op_type = 'Concrete07'

    def __init__(self, osi, fc, epsc, ec, ft, et, xp, xn, r):
        """
        Initial method for Concrete07

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc: float
            Concrete compressive strength (compression is negative)
        epsc: float
            Concrete strain at maximum compressive strength
        ec: float
            Initial elastic modulus of the concrete
        ft: float
            Tensile strength of concrete (tension is positive)
        et: float
            Tensile strain at max tensile strength of concrete
        xp: float
            Non-dimensional term that defines the strain at which the straight line descent begins in tension
        xn: float
            Non-dimensional term that defines the strain at which the straight line descent begins in compression
        r: float
            Parameter that controls the nonlinear descending branch

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.Concrete07(osi, fc=1.0, epsc=1.0, ec=1.0, ft=1.0, et=1.0, xp=1.0, xn=1.0, r=1.0)
        """
        self.osi = osi
        self.fc = float(fc)
        self.epsc = float(epsc)
        self.ec = float(ec)
        self.ft = float(ft)
        self.et = float(et)
        self.xp = float(xp)
        self.xn = float(xn)
        self.r = float(r)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc, self.epsc, self.ec, self.ft, self.et, self.xp, self.xn, self.r]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class Concrete01WithSITC(UniaxialMaterialBase):
    """
    The Concrete01WithSITC UniaxialMaterial Class
    
    This command is used to construct a modified uniaxial Kent-Scott-Park concrete material object with degraded linear
    unloading/reloading stiffness according to the work of Karsan-Jirsa and no tensile strength. The modification is to
    model the effect of Stuff In The Cracks (SITC).
    """
    op_type = 'Concrete01WithSITC'

    def __init__(self, osi, fpc, epsc0, fpcu, eps_u, end_strain_sitc=0.01):
        """
        Initial method for Concrete01WithSITC

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fpc: float
            Concrete compressive strength at 28 days (compression is negative)
        epsc0: float
            Concrete strain at maximum strength
        fpcu: float
            Concrete crushing strength
        eps_u: float
            Concrete strain at crushing strength
        end_strain_sitc: float, optional
            Optional, default = 0.03

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.Concrete01WithSITC(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0, end_strain_sitc=0.01)
        """
        self.osi = osi
        self.fpc = float(fpc)
        self.epsc0 = float(epsc0)
        self.fpcu = float(fpcu)
        self.eps_u = float(eps_u)
        self.end_strain_sitc = float(end_strain_sitc)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fpc, self.epsc0, self.fpcu, self.eps_u, self.end_strain_sitc]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class ConfinedConcrete01(UniaxialMaterialBase):
    """
    The ConfinedConcrete01 UniaxialMaterial Class
    
    
    """
    op_type = 'ConfinedConcrete01'

    def __init__(self, osi, sec_type, fpc, ec, epscu_type, epscu_val, nu, l1, l2, l3, phis, big_s, fyh, es0, ha_ratio, mu, phi_lon, internal_args: list=None, wrap_args: list=None, gravel=False, silica=False, tol: float=None, max_num_iter: int=None, epscu_limit: float=None, st_ratio=None):
        """
        Initial method for ConfinedConcrete01

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec_type: str
            Object for the transverse reinforcement configuration. see image below. * ``'s1'`` square section with s1
            type of transverse reinforcement with or without external frp wrapping * ``'s2'`` square section with s2 type of
            transverse reinforcement with or without external frp wrapping * ``'s3'`` square section with s3 type of
            transverse reinforcement with or without external frp wrapping * ``'s4a'`` square section with s4a type
            of transverse reinforcement with or without external frp wrapping * ``'s4b'`` square section with s4b
            type of transverse reinforcement with or without external frp wrapping * ``'s5'`` square section
            with s5 type of transverse reinforcement with or without external frp wrapping * ``'c'``
            circular section with or without external frp wrapping * ``'r'`` rectangular section
            with or without external frp wrapping.
        fpc: float
            Unconfined cylindrical strength of concrete specimen.
        ec: float
            Initial elastic modulus of unconfined concrete.
        epscu_type: str
            Method to define confined concrete ultimate strain * ``-epscu`` then value is confined concrete ultimate
            strain, * ``-gamma`` then value is the ratio of the strength corresponding to ultimate strain to the peak strength
            of the confined concrete in the range [0, epsculimit] then epsculimit (optional, default: 0.05) will be assumed
            as ultimate strain.
        epscu_val: float
            Value for the definition of the concrete ultimate strain
        nu: str
            Or |list|                definition for poisson's ratio. * ``*['-nu', <value of poisson's ratio>]`` *
            ``'-varub'`` poisson's ratio is defined as a function of axial strain by means of the expression proposed by
            braga et al. (2006) with the upper bound equal to 0.5 *``'-varnoub'`` poisson's ratio is defined as a
            function of axial strain by means of the expression proposed by braga et al. (2006) without any upper bound.
        l1: float
            Length/diameter of square/circular core section measured respect to the hoop center line.
        l2: float
            Additional dimensions when multiple hoops are being used.
        l3: float
            Additional dimensions when multiple hoops are being used.
        phis: float
            Hoop diameter. if section arrangement has multiple hoops it refers to the external hoop.
        big_s: float
            Hoop spacing.
        fyh: float
            Yielding strength of the hoop steel.
        es0: float
            Elastic modulus of the hoop steel.
        ha_ratio: float
            Hardening ratio of the hoop steel.
        mu: float
            Ductility factor of the hoop steel.
        phi_lon: float
            Diameter of longitudinal bars.
        internal_args: list, optional
            ``internalargs= [phisi, si, fyhi, es0i, haratioi, mui]`` optional parameters for defining the internal
            transverse reinforcement. if they are not specified they will be assumed equal to the external ones
        wrap_args: list, optional
            ``wrapargs=[cover, am, sw, ful, es0w]`` optional parameters required when section is strengthened with frp
            wraps. * ``cover`` cover thickness measured from the outer line of hoop. * ``am`` total area of frp wraps (number of
            layers x wrap thickness x wrap width). * ``sw`` spacing of frp wraps (if continuous wraps are used the spacing is
            equal to the wrap width). * ``ful`` ultimate strength of frp wraps. * ``es0w`` elastic modulus of frp wraps.
        gravel: bool
            Unknown
        silica: bool
            Unknown
        tol: float, optional
            Unknown
        max_num_iter: int, optional
            Unknown
        epscu_limit: float, optional
            Unknown
        st_ratio: unk, optional
            Unknown

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.ConfinedConcrete01(osi, sec_type='R', fpc=1.0, ec=1.0, epscu_type=1, epscu_val=1.0, nu=1, l1=1.0, l2=1.0, l3=1.0, phis=1.0, big_s=1.0, fyh=1.0, es0=1.0, ha_ratio=1.0, mu=1.0, phi_lon=1.0, gravel=1, tol=1.0, max_num_iter=1, epscu_limit=1.0, st_ratio=1)
        """
        self.osi = osi
        self.sec_type = sec_type
        self.fpc = float(fpc)
        self.ec = float(ec)
        self.epscu_type = epscu_type
        self.epscu_val = float(epscu_val)
        self.nu = nu
        self.l1 = float(l1)
        self.l2 = float(l2)
        self.l3 = float(l3)
        self.phis = float(phis)
        self.big_s = float(big_s)
        self.fyh = float(fyh)
        self.es0 = float(es0)
        self.ha_ratio = float(ha_ratio)
        self.mu = float(mu)
        self.phi_lon = float(phi_lon)
        self.internal_args = internal_args
        self.wrap_args = wrap_args
        self.gravel = gravel
        self.silica = silica
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        if max_num_iter is None:
            self.max_num_iter = None
        else:
            self.max_num_iter = int(max_num_iter)
        if epscu_limit is None:
            self.epscu_limit = None
        else:
            self.epscu_limit = float(epscu_limit)
        self.st_ratio = st_ratio
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.sec_type, self.fpc, self.ec, self.epscu_type, self.epscu_val, self.nu, self.l1, self.l2, self.l3, self.phis, self.big_s, self.fyh, self.es0, self.ha_ratio, self.mu, self.phi_lon]
        if getattr(self, 'internal_args') is not None:
            self._parameters += ['-internal', *self.internal_args]
        if getattr(self, 'wrap_args') is not None:
            self._parameters += ['-wrap', *self.wrap_args]
        if getattr(self, 'gravel'):
            self._parameters += ['-gravel']
        if getattr(self, 'silica'):
            self._parameters += ['-silica']
        if getattr(self, 'tol') is not None:
            self._parameters += ['-tol', self.tol]
        if getattr(self, 'max_num_iter') is not None:
            self._parameters += ['-maxNumIter', self.max_num_iter]
        if getattr(self, 'epscu_limit') is not None:
            self._parameters += ['-epscuLimit', self.epscu_limit]
        if getattr(self, 'st_ratio') is not None:
            self._parameters += ['-stRatio', self.st_ratio]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_fc(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'fc', value, ele, eles)

    def set_epsco(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'epsco', value, ele, eles)

    def set_fcu(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'fcu', value, ele, eles)

    def set_epscu(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'epscu', value, ele, eles)


class ConcreteD(UniaxialMaterialBase):
    """
    The ConcreteD UniaxialMaterial Class
    
    This command is used to construct a concrete material based on the Chinese design code.
    """
    op_type = 'ConcreteD'

    def __init__(self, osi, fc, epsc, ft, epst, ec, alphac, alphat, cesp=0.25, etap=1.15):
        """
        Initial method for ConcreteD

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc: float
            Concrete compressive strength
        epsc: float
            Concrete strain at corresponding to compressive strength
        ft: float
            Concrete tensile strength
        epst: float
            Concrete strain at corresponding to tensile strength
        ec: float
            Concrete initial elastic modulus
        alphac: float
            Compressive descending parameter
        alphat: float
            Tensile descending parameter
        cesp: float, optional
            Plastic parameter, recommended values: 0.2~0.3
        etap: float, optional
            Plastic parameter, recommended values: 1.0~1.3

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.ConcreteD(osi, fc=1.0, epsc=1.0, ft=1.0, epst=1.0, ec=1.0, alphac=1.0, alphat=1.0, cesp=0.25, etap=1.15)
        """
        self.osi = osi
        self.fc = float(fc)
        self.epsc = float(epsc)
        self.ft = float(ft)
        self.epst = float(epst)
        self.ec = float(ec)
        self.alphac = float(alphac)
        self.alphat = float(alphat)
        self.cesp = float(cesp)
        self.etap = float(etap)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc, self.epsc, self.ft, self.epst, self.ec, self.alphac, self.alphat, self.cesp, self.etap]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class FRPConfinedConcrete(UniaxialMaterialBase):
    """
    The FRPConfinedConcrete UniaxialMaterial Class
    
    This command is used to construct a uniaxial Megalooikonomou-Monti-Santini concrete material object with degraded
    linear unloading/reloading stiffness according to the work of Karsan-Jirsa and no tensile strength.
    """
    op_type = 'FRPConfinedConcrete'

    def __init__(self, osi, fpc1, fpc2, epsc0, big_d, c, ej, sj, tj, eju, big_s, fyl, fyh, dlong, dtrans, es, nu0, k, use_buck):
        """
        Initial method for FRPConfinedConcrete

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fpc1: float
            Concrete core compressive strength.
        fpc2: float
            Concrete cover compressive strength.
        epsc0: float
            Strain corresponding to unconfined concrete strength.
        big_d: float
            Diameter of the circular section.
        c: float
            Dimension of concrete cover (until the outer edge of steel stirrups)
        ej: float
            Elastic modulus of the fiber reinforced polymer (frp) jacket.
        sj: float
            Clear spacing of the frp strips - zero if frp jacket is continuous.
        tj: float
            Total thickness of the frp jacket.
        eju: float
            Rupture strain of the frp jacket from tensile coupons.
        big_s: float
            Spacing of the steel spiral/stirrups.
        fyl: float
            Yielding strength of longitudinal steel bars.
        fyh: float
            Yielding strength of the steel spiral/stirrups.
        dlong: float
            Diameter of the longitudinal bars of the circular section.
        dtrans: float
            Diameter of the steel spiral/stirrups.
        es: float
            Elastic modulus of steel.
        nu0: float
            Initial poisson's coefficient for concrete.
        k: float
            Reduction factor for the rupture strain of the frp jacket, recommended values 0.5-0.8.
        use_buck: float
            Frp jacket failure criterion due to buckling of longitudinal compressive steel bars (0 = not                
                                       include it, 1= to include it).

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.FRPConfinedConcrete(osi, fpc1=1.0, fpc2=1.0, epsc0=1.0, big_d=1.0, c=1.0, ej=1.0, sj=1.0, tj=1.0, eju=1.0, big_s=1.0, fyl=1.0, fyh=1.0, dlong=1.0, dtrans=1.0, es=1.0, nu0=1.0, k=1.0, use_buck=1.0)
        """
        self.osi = osi
        self.fpc1 = float(fpc1)
        self.fpc2 = float(fpc2)
        self.epsc0 = float(epsc0)
        self.big_d = float(big_d)
        self.c = float(c)
        self.ej = float(ej)
        self.sj = float(sj)
        self.tj = float(tj)
        self.eju = float(eju)
        self.big_s = float(big_s)
        self.fyl = float(fyl)
        self.fyh = float(fyh)
        self.dlong = float(dlong)
        self.dtrans = float(dtrans)
        self.es = float(es)
        self.nu0 = float(nu0)
        self.k = float(k)
        self.use_buck = float(use_buck)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fpc1, self.fpc2, self.epsc0, self.big_d, self.c, self.ej, self.sj, self.tj, self.eju, self.big_s, self.fyl, self.fyh, self.dlong, self.dtrans, self.es, self.nu0, self.k, self.use_buck]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_fc1(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'fc1', value, ele, eles)

    def set_fc2(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'fc2', value, ele, eles)

    def set_epsco(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'epsco', value, ele, eles)

    def set_d(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'D', value, ele, eles)

    def set_c(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'c', value, ele, eles)

    def set_ej(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Ej', value, ele, eles)

    def set_sj(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Sj', value, ele, eles)

    def set_tj(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'tj', value, ele, eles)

    def set_eju(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'eju', value, ele, eles)

    def set_s(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'S', value, ele, eles)

    def set_fyl(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'fyl', value, ele, eles)

    def set_fyh(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'fyh', value, ele, eles)

    def set_dlong(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'dlong', value, ele, eles)

    def set_dtrans(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'dtrans', value, ele, eles)

    def set_es(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'Es', value, ele, eles)

    def set_vo(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'vo', value, ele, eles)

    def set_k(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'k', value, ele, eles)

    def set_use_buck(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'useBuck', value, ele, eles)


class FRPConfinedConcrete02JacketC(UniaxialMaterialBase):
    """
    The FRPConfinedConcrete02JacketC UniaxialMaterial Class
    
    | DEVELOPED AND IMPLEMENTED BY:| Jin-Yu LU, Southeast University, Nanjing, China| Guan LIN (guanlin@polyu.edu.hk),
    Hong Kong Polytechnic University, Hong Kong, China.

    Figure 1 Hysteretic Stress-Strain Relation.. image::
    /_static/FRPConfinedConcrete02/Figure1.pngThis command is used to construct a uniaxial hysteretic
    stress-strain model for fiber-reinforced polymer (FRP)-confined concrete. The envelope
    compressive stress-strain response is described by a parabolic first portion and a
    linear second portion with smooth connection between them (Figure 1). The
    hysteretic rules of compression are based on Lam and Teng’s (2009)
    model. The cyclic linear tension model of Yassin (1994) for
    unconfined concrete (as adopted in Concrete02) is used
    with slight modifications to describe the tensile behavior of FRP-confined concrete (Teng et al. 2015).
    """
    op_type = 'FRPConfinedConcrete02'

    def __init__(self, osi, fc0, ec, ec0, tfrp, efrp, erup, big_r):
        """
        Initial method for FRPConfinedConcrete02JacketC

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc0: float
            Compressive strength of unconfined concrete (compression is negative)
        ec: float
            Elastic modulus of unconfined concrete (=4730√(-$fc0(mpa)))
        ec0: float
            Axial strain corresponding to unconfined concrete strength (≈ 0.002)
        tfrp: float
            Thickness of an frp jacket
        efrp: float
            Tensile elastic modulus of an frp jacket
        erup: float
            Hoop rupture strain of an frp jacket
        big_r: float
            Radius of circular column section
        """
        self.osi = osi
        self.fc0 = float(fc0)
        self.ec = float(ec)
        self.ec0 = float(ec0)
        self.tfrp = float(tfrp)
        self.efrp = float(efrp)
        self.erup = float(erup)
        self.big_r = float(big_r)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc0, self.ec, self.ec0, '-JacketC', self.tfrp, self.efrp, self.erup, self.big_r]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

class FRPConfinedConcrete02Ultimate(UniaxialMaterialBase):
    """
    The FRPConfinedConcrete02Ultimate UniaxialMaterial Class
    
    | DEVELOPED AND IMPLEMENTED BY:| Jin-Yu LU, Southeast University, Nanjing, China| Guan LIN (guanlin@polyu.edu.hk),
    Hong Kong Polytechnic University, Hong Kong, China.

    Figure 1 Hysteretic Stress-Strain Relation.. image::
    /_static/FRPConfinedConcrete02/Figure1.pngThis command is used to construct a uniaxial hysteretic
    stress-strain model for fiber-reinforced polymer (FRP)-confined concrete. The envelope
    compressive stress-strain response is described by a parabolic first portion and a
    linear second portion with smooth connection between them (Figure 1). The
    hysteretic rules of compression are based on Lam and Teng’s (2009)
    model. The cyclic linear tension model of Yassin (1994) for
    unconfined concrete (as adopted in Concrete02) is used
    with slight modifications to describe the tensile behavior of FRP-confined concrete (Teng et al. 2015).
    """
    op_type = 'FRPConfinedConcrete02'

    def __init__(self, osi, fc0, ec, ec0, fcu, ecu, ft, ets, unit):
        """
        Initial method for FRPConfinedConcrete02Ultimate

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc0: float
            Compressive strength of unconfined concrete (compression is negative)
        ec: float
            Elastic modulus of unconfined concrete (=4730√(-$fc0(mpa)))
        ec0: float
            Axial strain corresponding to unconfined concrete strength (≈ 0.002)
        fcu: float
            Ultimate stress of frp-confined concrete ($fcu ≥ $fc0)
        ecu: float
            Ultimate strain of frp-confined concrete
        ft: float
            Tensile strength of unconfined concrete (=0.632√(-$fc0(mpa)))
        ets: float
            Stiffness of tensile softening (≈ 0.05 ec)
        unit: float
            Unit indicator, unit = 1 for si metric units; unit = 0 for us customary units
        """
        self.osi = osi
        self.fc0 = float(fc0)
        self.ec = float(ec)
        self.ec0 = float(ec0)
        self.fcu = float(fcu)
        self.ecu = float(ecu)
        self.ft = float(ft)
        self.ets = float(ets)
        self.unit = float(unit)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc0, self.ec, self.ec0, '-Ultimate', self.fcu, self.ecu, self.ft, self.ets, self.unit]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class ConcreteCM(UniaxialMaterialBase):
    """
    The ConcreteCM UniaxialMaterial Class
    
    This command is used to construct a uniaxialMaterial ConcreteCM (Kolozvari et al., 2015), which is a uniaxial
    hysteretic constitutive model for concrete developed by Chang and Mander (1994).
    """
    op_type = 'ConcreteCM'

    def __init__(self, osi, fpcc, epcc, ec, rc, xcrn, ft, et, rt, xcrp, gap_close: float=None):
        r"""
        Initial method for ConcreteCM

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fpcc: float
            Compressive strength (:math:`f'_c`)
        epcc: float
            Strain at compressive strength (:math:`\epsilon'_c`)
        ec: float
            Initial tangent modulus (:math:`e_c`)
        rc: float
            Shape parameter in tsai's equation defined for compression (:math:`r_c`)
        xcrn: float
            Non-dimensional critical strain on compression envelope (:math:`\epsilon^{-}_{cr}`, where the envelope curve
            starts following a straight line)
        ft: float
            Tensile strength (:math:`f_t`)
        et: float
            Strain at tensile strength (:math:`\epsilon_t`)
        rt: float
            Shape parameter in tsai's equation defined for tension (:math:`r_t`)
        xcrp: float
            Non-dimensional critical strain on tension envelope (:math:`\epsilon^{+}_{cr}`, where the envelope curve
            starts following a straight line - large value [e.g., 10000] recommended when tension stiffening is considered)
        gap_close: float, optional
            Gapclose = 0, less gradual gap closure (default); gapclose = 1, more gradual gap closure

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.uniaxial_material.ConcreteCM(osi, fpcc=1.0, epcc=1.0, ec=1.0, rc=1.0, xcrn=1.0, ft=1.0, et=1.0, rt=1.0, xcrp=1.0, gap_close=0)
        """
        self.osi = osi
        self.fpcc = float(fpcc)
        self.epcc = float(epcc)
        self.ec = float(ec)
        self.rc = float(rc)
        self.xcrn = float(xcrn)
        self.ft = float(ft)
        self.et = float(et)
        self.rt = float(rt)
        self.xcrp = float(xcrp)
        if gap_close is None:
            self.gap_close = None
        else:
            self.gap_close = float(gap_close)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fpcc, self.epcc, self.ec, self.rc, self.xcrn, self.ft, self.et, self.rt, self.xcrp]
        if getattr(self, 'gap_close') is not None:
            self._parameters += ['-GapClose', self.gap_close]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class TDConcrete(UniaxialMaterialBase):
    """
    The TDConcrete UniaxialMaterial Class
    
    This command is used to construct a uniaxial time-dependent concrete material object with linear behavior in
    compression, nonlinear behavior in tension (REF: Tamai et al., 1988) and creep and shrinkage according to ACI 209R-92.
    """
    op_type = 'TDConcrete'

    def __init__(self, osi, fc, fct, ec, beta, t_d, epsshu, psish, tcr, phiu, psicr1, psicr2, tcast):
        """
        Initial method for TDConcrete

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc: float
            Concrete compressive strength (compression is negative)
        fct: float
            Concrete tensile strength (tension is positive)
        ec: float
            Concrete modulus of elasticity
        beta: float
            Tension softening parameter (tension softening exponent)
        t_d: float
            Analysis time at initiation of drying (in days)
        epsshu: float
            Ultimate shrinkage strain as per aci 209r-92 (shrinkage is negative)
        psish: float
            Fitting parameter of the shrinkage time evolution function as per aci 209r-92
        tcr: float
            Creep model age (in days)
        phiu: float
            Ultimate creep coefficient as per aci 209r-92
        psicr1: float
            Fitting parameter of the creep time evolution function as per aci 209r-92
        psicr2: float
            Fitting parameter of the creep time evolution function as per aci 209r-92
        tcast: float
            Analysis time corresponding to concrete casting (in days; minimum value 2.0)
        """
        self.osi = osi
        self.fc = float(fc)
        self.fct = float(fct)
        self.ec = float(ec)
        self.beta = float(beta)
        self.t_d = float(t_d)
        self.epsshu = float(epsshu)
        self.psish = float(psish)
        self.tcr = float(tcr)
        self.phiu = float(phiu)
        self.psicr1 = float(psicr1)
        self.psicr2 = float(psicr2)
        self.tcast = float(tcast)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc, self.fct, self.ec, self.beta, self.t_d, self.epsshu, self.psish, self.tcr, self.phiu, self.psicr1, self.psicr2, self.tcast]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class TDConcreteEXP(UniaxialMaterialBase):
    """
    The TDConcreteEXP UniaxialMaterial Class
    
    This command is used to construct a uniaxial time-dependent concrete material object with linear behavior in
    compression, nonlinear behavior in tension (REF: Tamai et al., 1988) and creep and shrinkage according to ACI 209R-92.
    """
    op_type = 'TDConcreteEXP'

    def __init__(self, osi, fc, fct, ec, beta, t_d, epsshu, psish, tcr, epscru, sig_cr, psicr1, psicr2, tcast):
        """
        Initial method for TDConcreteEXP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc: float
            Concrete compressive strength (compression is negative)
        fct: float
            Concrete tensile strength (tension is positive)
        ec: float
            Concrete modulus of elasticity
        beta: float
            Tension softening parameter (tension softening exponent)
        t_d: float
            Analysis time at initiation of drying (in days)
        epsshu: float
            Ultimate shrinkage strain as per aci 209r-92 (shrinkage is negative)
        psish: float
            Fitting parameter of the shrinkage time evolution function as per aci 209r-92
        tcr: float
            Creep model age (in days)
        epscru: float
            Ultimate creep strain (e.g., taken from experimental measurements)
        sig_cr: float
            Concrete compressive stress (input as negative) associated with $epscru (e.g., experimentally applied)
        psicr1: float
            Fitting parameter of the creep time evolution function as per aci 209r-92
        psicr2: float
            Fitting parameter of the creep time evolution function as per aci 209r-92
        tcast: float
            Analysis time corresponding to concrete casting (in days; minimum value 2.0)
        """
        self.osi = osi
        self.fc = float(fc)
        self.fct = float(fct)
        self.ec = float(ec)
        self.beta = float(beta)
        self.t_d = float(t_d)
        self.epsshu = float(epsshu)
        self.psish = float(psish)
        self.tcr = float(tcr)
        self.epscru = float(epscru)
        self.sig_cr = float(sig_cr)
        self.psicr1 = float(psicr1)
        self.psicr2 = float(psicr2)
        self.tcast = float(tcast)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc, self.fct, self.ec, self.beta, self.t_d, self.epsshu, self.psish, self.tcr, self.epscru, self.sig_cr, self.psicr1, self.psicr2, self.tcast]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class TDConcreteMC10(UniaxialMaterialBase):
    """
    The TDConcreteMC10 UniaxialMaterial Class
    
    This command is used to construct a uniaxial time-dependent concrete material object with linear behavior in
    compression, nonlinear behavior in tension (REF: Tamai et al., 1988) and creep and shrinkage according to fib
    Model Code 2010.
    """
    op_type = 'TDConcreteMC10'

    def __init__(self, osi, fc, fct, ec, ecm, beta, t_d, epsba, epsbb, epsda, epsdb, phiba, phibb, phida, phidb, tcast, cem):
        """
        Initial method for TDConcreteMC10

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc: float
            Concrete compressive strength (compression is negative)
        fct: float
            Concrete tensile strength (tension is positive)
        ec: float
            Concrete modulus of elasticity at loading age
        ecm: float
            Concrete modulus of elasticity at 28 days
        beta: float
            Tension softening parameter (tension softening exponent)
        t_d: float
            Analysis time at initiation of drying (in days)
        epsba: float
            Ultimate basic shrinkage strain (input as negative) as per fib model code 2010
        epsbb: float
            Fitting parameter of the basic shrinkage time evolution function as per fib model code 2010
        epsda: float
            Product of ultimate drying shrinkage strain and relative humidity function as per fib model code 2010
        epsdb: float
            Fitting parameter of the basic shrinkage time evolution function as per fib model code 2010
        phiba: float
            Parameter for the effect of compressive strength on basic creep as per fib model code 2010
        phibb: float
            Fitting parameter of the basic creep time evolution function as per fib model code 2010
        phida: float
            Product of the effect of compressive strength and relative humidity on drying creep as per fib model code
            2010
        phidb: float
            Fitting parameter of the drying creep time evolution function as per fib model code 2010
        tcast: float
            Analysis time corresponding to concrete casting (in days; minimum value 2.0)
        cem: float
            Coefficient dependent on the type of cement as per fib model code 2010
        """
        self.osi = osi
        self.fc = float(fc)
        self.fct = float(fct)
        self.ec = float(ec)
        self.ecm = float(ecm)
        self.beta = float(beta)
        self.t_d = float(t_d)
        self.epsba = float(epsba)
        self.epsbb = float(epsbb)
        self.epsda = float(epsda)
        self.epsdb = float(epsdb)
        self.phiba = float(phiba)
        self.phibb = float(phibb)
        self.phida = float(phida)
        self.phidb = float(phidb)
        self.tcast = float(tcast)
        self.cem = float(cem)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc, self.fct, self.ec, self.ecm, self.beta, self.t_d, self.epsba, self.epsbb, self.epsda, self.epsdb, self.phiba, self.phibb, self.phida, self.phidb, self.tcast, self.cem]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class TDConcreteMC10NL(UniaxialMaterialBase):
    """
    The TDConcreteMC10NL UniaxialMaterial Class
    
    This command is used to construct a uniaxial time-dependent concrete material object with non-linear behavior in
    compression (REF: Concrete02), nonlinear behavior in tension (REF: Tamai et al., 1988) and creep and shrinkage
    according to fib Model Code 2010.
    """
    op_type = 'TDConcreteMC10NL'

    def __init__(self, osi, fc, fcu, epscu, fct, ec, ecm, beta, t_d, epsba, epsbb, epsda, epsdb, phiba, phibb, phida, phidb, tcast, cem):
        """
        Initial method for TDConcreteMC10NL

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        fc: float
            Concrete compressive strength (compression is negative)
        fcu: float
            Concrete crushing strength (compression is negative)
        epscu: float
            Concrete strain at crushing strength (input as negative)
        fct: float
            Concrete tensile strength (tension is positive)
        ec: float
            Concrete modulus of elasticity at loading age
        ecm: float
            Concrete modulus of elasticity at 28 days
        beta: float
            Tension softening parameter (tension softening exponent)
        t_d: float
            Analysis time at initiation of drying (in days)
        epsba: float
            Ultimate basic shrinkage strain (input as negative) as per fib model code 2010
        epsbb: float
            Fitting parameter of the basic shrinkage time evolution function as per fib model code 2010
        epsda: float
            Product of ultimate drying shrinkage strain and relative humidity function as per fib model code 2010
        epsdb: float
            Fitting parameter of the basic shrinkage time evolution function as per fib model code 2010
        phiba: float
            Parameter for the effect of compressive strength on basic creep as per fib model code 2010
        phibb: float
            Fitting parameter of the basic creep time evolution function as per fib model code 2010
        phida: float
            Product of the effect of compressive strength and relative humidity on drying creep as per fib model code
            2010
        phidb: float
            Fitting parameter of the drying creep time evolution function as per fib model code 2010
        tcast: float
            Analysis time corresponding to concrete casting (in days; minimum value 2.0)
        cem: float
            Coefficient dependent on the type of cement as per fib model code 2010
        """
        self.osi = osi
        self.fc = float(fc)
        self.fcu = float(fcu)
        self.epscu = float(epscu)
        self.fct = float(fct)
        self.ec = float(ec)
        self.ecm = float(ecm)
        self.beta = float(beta)
        self.t_d = float(t_d)
        self.epsba = float(epsba)
        self.epsbb = float(epsbb)
        self.epsda = float(epsda)
        self.epsdb = float(epsdb)
        self.phiba = float(phiba)
        self.phibb = float(phibb)
        self.phida = float(phida)
        self.phidb = float(phidb)
        self.tcast = float(tcast)
        self.cem = float(cem)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.fc, self.fcu, self.epscu, self.fct, self.ec, self.ecm, self.beta, self.t_d, self.epsba, self.epsbb, self.epsda, self.epsdb, self.phiba, self.phibb, self.phida, self.phidb, self.tcast, self.cem]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)



from o3seespy.command.element.base_element import ElementBase


class ElasticBeamColumn2D(ElementBase):
    """
    The ElasticBeamColumn2D Element Class
    
    This command is used to construct an elasticBeamColumn element object. The arguments for the construction of an
    elastic beam-column element depend on the dimension of the problem, (ndm)

    For a two-dimensional problem
    """
    op_type = 'elasticBeamColumn'

    def __init__(self, osi, ele_nodes, area, e_mod, iz, transf, mass: float=None, c_mass=False, release_code=None):
        """
        Initial method for ElasticBeamColumn2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        area: float
            Cross-sectional area of element
        e_mod: float
            Young's modulus
        iz: float
            Second moment of area about the local z-axis
        transf: obj
            Identifier for previously-defined coordinate-transformation (crdtransf) object
        mass: float, optional
            Element mass per unit length (optional, default = 0.0)
        c_mass: bool
            To form consistent mass matrix (optional, default = lumped mass matrix)
        release_code: None, optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> transf = o3.geom_transf.Linear2D(osi, [])
        >>> o3.element.ElasticBeamColumn2D(osi, ele_nodes=ele_nodes, area=1.0, e_mod=1.0, iz=1.0, transf=transf, mass=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.area = float(area)
        self.e_mod = float(e_mod)
        self.iz = float(iz)
        self.transf = transf
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        self.c_mass = c_mass
        self.release_code = release_code
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.area, self.e_mod, self.iz, self.transf.tag]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'c_mass'):
            self._parameters += ['-cMass']
        if getattr(self, 'release_code') is not None:
            self._parameters += ['-release', self.release_code]
        self.to_process(osi)


class ElasticBeamColumn3D(ElementBase):
    """
    The ElasticBeamColumn3D Element Class
    
    This command is used to construct an elasticBeamColumn element object. The arguments for the construction of an
    elastic beam-column element depend on the dimension of the problem, (ndm)

    For a three-dimensional problem
    """
    op_type = 'elasticBeamColumn'

    def __init__(self, osi, ele_nodes, area, e_mod, g_mod, jxx, iy, iz, transf, mass: float=None, c_mass=False):
        """
        Initial method for ElasticBeamColumn3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        area: float
            Cross-sectional area of element
        e_mod: float
            Young's modulus
        g_mod: float
            Shear modulus
        jxx: float
            Torsional moment of inertia of cross section
        iy: float
            Second moment of area about the local y-axis
        iz: float
            Second moment of area about the local z-axis
        transf: obj
            Identifier for previously-defined coordinate-transformation (crdtransf) object
        mass: float, optional
            Element mass per unit length (optional, default = 0.0)
        c_mass: bool
            To form consistent mass matrix (optional, default = lumped mass matrix)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> transf = o3.geom_transf.Linear2D(osi, [])
        >>> o3.element.ElasticBeamColumn3D(osi, ele_nodes=ele_nodes, area=1.0, e_mod=1.0, g_mod=1.0, jxx=1.0, iy=1.0, iz=1.0, transf=transf, mass=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.area = float(area)
        self.e_mod = float(e_mod)
        self.g_mod = float(g_mod)
        self.jxx = float(jxx)
        self.iy = float(iy)
        self.iz = float(iz)
        self.transf = transf
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        self.c_mass = c_mass
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.area, self.e_mod, self.g_mod, self.jxx, self.iy, self.iz, self.transf.tag]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'c_mass'):
            self._parameters += ['-cMass']
        self.to_process(osi)


class ModElasticBeam2D(ElementBase):
    """
    The ModElasticBeam2D Element Class
    
    This command is used to construct a ModElasticBeam2d element object. The arguments for the construction of an
    elastic beam-column element with stiffness modifiers is applicable for 2-D problems. This element should be used
    for modelling of a structural element with an equivalent combination of one elastic element with
    stiffness-proportional damping, and two springs at its two ends with no stiffness proportional
    damping to represent a prismatic section. The modelling technique is based on a number of
    analytical studies discussed in Zareian and Medina (2010) and Zareian and Krawinkler
    (2009) and is utilized in order to solve problems related to numerical damping in
    dynamic analysis of frame structures with concentrated plasticity springs.

    
    """
    op_type = 'ModElasticBeam2d'

    def __init__(self, osi, ele_nodes, area, e_mod, iz, k11, k33, k44, transf, c_mass=False, mass: float=None):
        """
        Initial method for ModElasticBeam2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        area: float
            Cross-sectional area of element
        e_mod: float
            Young's modulus
        iz: float
            Second moment of area about the local z-axis
        k11: float
            Stiffness modifier for translation
        k33: float
            Stiffness modifier for translation
        k44: float
            Stiffness modifier for rotation
        transf: obj
            Identifier for previously-defined coordinate-transformation (crdtransf) object
        c_mass: bool
            To form consistent mass matrix (optional, default = lumped mass matrix)
        mass: float, optional
            Element mass per unit length (optional, default = 0.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.area = float(area)
        self.e_mod = float(e_mod)
        self.iz = float(iz)
        self.k11 = float(k11)
        self.k33 = float(k33)
        self.k44 = float(k44)
        self.transf = transf
        self.c_mass = c_mass
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.area, self.e_mod, self.iz, self.k11, self.k33, self.k44, self.transf.tag]
        if getattr(self, 'c_mass'):
            self._parameters += ['-cMass']
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)


class ElasticTimoshenkoBeam2D(ElementBase):
    """
    The ElasticTimoshenkoBeam2D Element Class
    
    This command is used to construct an ElasticTimoshenkoBeam element object. A Timoshenko beam is a frame member that
    accounts for shear deformations. The arguments for the construction of an elastic Timoshenko beam element depend on
    the dimension of the problem, ndm:

    For a two-dimensional problem:
    """
    op_type = 'ElasticTimoshenkoBeam'

    def __init__(self, osi, ele_nodes, e_mod, g_mod, area, iz, avy, transf, c_mass=False, mass: float=None):
        """
        Initial method for ElasticTimoshenkoBeam2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        e_mod: float
            Young's modulus
        g_mod: float
            Shear modulus
        area: float
            Cross-sectional area of element
        iz: float
            Second moment of area about the local z-axis
        avy: float
            Shear area for the local y-axis
        transf: obj
            Identifier for previously-defined coordinate-transformation (crdtransf) object
        c_mass: bool
            To form consistent mass matrix (optional, default = lumped mass matrix)
        mass: float, optional
            Element mass per unit length (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> transf = o3.geom_transf.Linear2D(osi, [])
        >>> o3.element.ElasticTimoshenkoBeam2D(osi, ele_nodes=ele_nodes, e_mod=1.0, g_mod=1.0, area=1.0, iz=1.0, avy=1.0, transf=transf, mass=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.e_mod = float(e_mod)
        self.g_mod = float(g_mod)
        self.area = float(area)
        self.iz = float(iz)
        self.avy = float(avy)
        self.transf = transf
        self.c_mass = c_mass
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.e_mod, self.g_mod, self.area, self.iz, self.avy, self.transf.tag]
        if getattr(self, 'c_mass'):
            self._parameters += ['-cMass']
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)


class ElasticTimoshenkoBeam3D(ElementBase):
    """
    The ElasticTimoshenkoBeam3D Element Class
    
    This command is used to construct an ElasticTimoshenkoBeam element object. A Timoshenko beam is a frame member that
    accounts for shear deformations. The arguments for the construction of an elastic Timoshenko beam element depend on
    the dimension of the problem, ndm:

    For a three-dimensional problem:
    """
    op_type = 'ElasticTimoshenkoBeam'

    def __init__(self, osi, ele_nodes, e_mod, g_mod, area, iz, jxx, iy, iz_2, avy, avz, transf, c_mass=False, mass: float=None):
        """
        Initial method for ElasticTimoshenkoBeam3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        e_mod: float
            Young's modulus
        g_mod: float
            Shear modulus
        area: float
            Cross-sectional area of element
        iz: float
            Second moment of area about the local z-axis
        jxx: float
            Torsional moment of inertia of cross section
        iy: float
            Second moment of area about the local y-axis
        iz_2: None
            
        avy: float
            Shear area for the local y-axis
        avz: float
            Shear area for the local z-axis
        transf: obj
            Identifier for previously-defined coordinate-transformation (crdtransf) object
        c_mass: bool
            To form consistent mass matrix (optional, default = lumped mass matrix)
        mass: float, optional
            Element mass per unit length (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> transf = o3.geom_transf.Linear2D(osi, [])
        >>> o3.element.ElasticTimoshenkoBeam3D(osi, ele_nodes=ele_nodes, e_mod=1.0, g_mod=1.0, area=1.0, iz=1.0, jxx=1.0, iy=1.0, iz_2=1, avy=1.0, avz=1.0, transf=transf, mass=1.0, c_mass=True)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.e_mod = float(e_mod)
        self.g_mod = float(g_mod)
        self.area = float(area)
        self.iz = float(iz)
        self.jxx = float(jxx)
        self.iy = float(iy)
        self.iz_2 = iz_2
        self.avy = float(avy)
        self.avz = float(avz)
        self.transf = transf
        self.c_mass = c_mass
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.e_mod, self.g_mod, self.area, self.iz, self.jxx, self.iy, self.iz_2, self.avy, self.avz, self.transf.tag]
        if getattr(self, 'c_mass'):
            self._parameters += ['-cMass']
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)



class DispBeamColumn(ElementBase):
    """
    The DispBeamColumn Element Class
    
    Create a dispBeamColumn element.
    """
    op_type = 'dispBeamColumn'

    def __init__(self, osi, ele_nodes, transf, integration, c_mass=False, mass: float=None):
        """
        Initial method for DispBeamColumn

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            List of two node objects
        transf: obj
            Object of transformation
        integration: obj
            Object of :func:`beamintegration`
        c_mass: None
            
        mass: float, optional
            Element mass density (per unit length), from which a lumped-mass matrix is formed 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0)
        >>> transf = o3.geom_transf.Linear2D(osi, [])
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> integration = o3.beam_integration.Lobatto(osi, sec, 5)
        >>> o3.element.DispBeamColumn(osi, ele_nodes=[i_node, j_node], transf=transf, integration=integration, c_mass=1, mass=0.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.transf = transf
        self.integration = integration
        self.c_mass = c_mass
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.transf.tag, self.integration.tag]
        if getattr(self, 'c_mass'):
            self._parameters += ['-cMass']
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)


class ForceBeamColumn(ElementBase):
    """
    The ForceBeamColumn Element Class
    
    Create a ForceBeamColumn element.
    """
    op_type = 'forceBeamColumn'

    def __init__(self, osi, ele_nodes, transf, integration, max_iter: int=None, tol: float=None, mass: float=None):
        """
        Initial method for ForceBeamColumn

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        transf: obj
            Object of transformation
        integration: obj
            Object of :func:`beamintegration`
        max_iter: int, optional
            Maximum number of iterations to undertake to satisfy element compatibility 
        tol: float, optional
            Tolerance for satisfaction of element compatibility 
        mass: float, optional
            Element mass density (per unit length), from which a lumped-mass matrix is formed 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0)
        >>> transf = o3.geom_transf.Linear2D(osi, [])
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> integration = o3.beam_integration.Lobatto(osi, sec, 5)
        >>> o3.element.ForceBeamColumn(osi, ele_nodes=[i_node, j_node], transf=transf, integration=integration, max_iter=10, tol=1e-12, mass=0.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.transf = transf
        self.integration = integration
        if max_iter is None:
            self.max_iter = None
        else:
            self.max_iter = int(max_iter)
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.transf.tag, self.integration.tag]
        if getattr(self, 'max_iter') is not None:
            self._parameters += ['-iter', self.max_iter]
        if getattr(self, 'tol') is not None:
            if getattr(self, 'max_iter') is None:
                raise ValueError('Cannot set: tol and not: max_iter')
            self._parameters += [self.tol]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)


class NonlinearBeamColumn(ElementBase):
    """
    The NonlinearBeamColumn Element Class
    
    Create a nonlinearBeamColumn element. This element is for backward compatability.
    """
    op_type = 'nonlinearBeamColumn'

    def __init__(self, osi, ele_nodes, num_intgr_pts, sec, transf, max_iter: int=None, tol: float=None, mass: float=None, int_type: str=None):
        """
        Initial method for NonlinearBeamColumn

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        num_intgr_pts: int
            Number of integration points.
        sec: obj
            Object of section
        transf: obj
            Object of transformation
        max_iter: int, optional
            Maximum number of iterations to undertake to satisfy element compatibility 
        tol: float, optional
            Tolerance for satisfaction of element compatibility 
        mass: float, optional
            Element mass density (per unit length), from which a lumped-mass matrix is formed 
        int_type: str, optional
            Integration type (optional, default is ``'lobatto'``) * ``'lobatto'`` * ``'legendre'`` * ``'radau'`` *
            ``'newtoncotes'`` * ``'trapezoidal'``

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> i_node = o3.node.Node(osi, 0.0, 0.0)
        >>> j_node = o3.node.Node(osi, 0.0, 1.0)
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> transf = o3.geom_transf.Linear2D(osi, [])
        >>> o3.element.NonlinearBeamColumn(osi, ele_nodes=[i_node, j_node], num_intgr_pts=1, sec=sec, transf=transf,
        >>>                                max_iter=10, tol=1e-12, mass=0.0, int_type="radau")
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.num_intgr_pts = int(num_intgr_pts)
        self.sec = sec
        self.transf = transf
        if max_iter is None:
            self.max_iter = None
        else:
            self.max_iter = int(max_iter)
        if tol is None:
            self.tol = None
        else:
            self.tol = float(tol)
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        self.int_type = int_type
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.num_intgr_pts, self.sec.tag, self.transf.tag]
        if getattr(self, 'max_iter') is not None:
            self._parameters += ['-iter', self.max_iter]
        if getattr(self, 'tol') is not None:
            if getattr(self, 'max_iter') is None:
                raise ValueError('Cannot set: tol and not: max_iter')
            self._parameters += [self.tol]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        if getattr(self, 'int_type') is not None:
            self._parameters += ['-integration', self.int_type]
        self.to_process(osi)


class DispBeamColumnInt(ElementBase):
    """
    The DispBeamColumnInt Element Class
    
    This command is used to construct a dispBeamColumnInt element object, which is a distributed-plasticity,
    displacement-based beam-column element which includes interaction between flexural and shear components.

    
    """
    op_type = 'dispBeamColumnInt'

    def __init__(self, osi, ele_nodes, num_intgr_pts, sec, transf, c_rot, mass: float=None):
        """
        Initial method for DispBeamColumnInt

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        num_intgr_pts: int
            Number of integration points along the element.
        sec: obj
            Identifier for previously-defined section object
        transf: obj
            Identifier for previously-defined coordinate-transformation (crdtransf) object
        c_rot: float
            Identifier for element center of rotation (or center of curvature distribution). fraction of the height
            distance from bottom to the center of rotation (0 to 1)
        mass: float, optional
            Element mass density (per unit length), from which a lumped-mass matrix is formed (optional, default=0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> transf = o3.geom_transf.Linear2D(osi, [])
        >>> o3.element.DispBeamColumnInt(osi, ele_nodes=ele_nodes, num_intgr_pts=4, sec=sec, transf=transf, c_rot=1.0, mass=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.num_intgr_pts = int(num_intgr_pts)
        self.sec = sec
        self.transf = transf
        self.c_rot = float(c_rot)
        if mass is None:
            self.mass = None
        else:
            self.mass = float(mass)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.num_intgr_pts, self.sec.tag, self.transf.tag, self.c_rot]
        if getattr(self, 'mass') is not None:
            self._parameters += ['-mass', self.mass]
        self.to_process(osi)


class MVLEM(ElementBase):
    """
    The MVLEM Element Class
    
    The MVLEM element command is used to generate a two-dimensional Multiple-Vertical-Line-Element-Model (MVLEM; Vulcano
    et al., 1988; Orakcal et al., 2004, Kolozvari et al., 2015) for simulation of flexure-dominated RC wall behavior. A
    single model element incorporates six global degrees of freedom, three of each located at the center of rigid top
    and bottom beams, as illustrated in Figure 1a. The axial/flexural response of the MVLEM is simulated by a series
    of uniaxial elements (or macro-fibers) connected to the rigid beams at the top and bottom (e.g., floor) levels,
    whereas the shear response is described by a shear spring located at height ch from the bottom of the wall
    element (Figure 1a). Shear and flexural responses of the model element are uncoupled. The relative
    rotation between top and bottom faces of the wall element occurs about the point located on the
    central axis of the element at height ch (Figure 1b). Rotations and resulting transverse
    displacements are calculated based on the wall curvature, derived from section and
    material properties, corresponding to the bending moment at height ch of each
    element (Figure 1b). A value of c=0.4 was recommended by Vulcano et al.
    (1988) based on comparison of the model response with experimental results.

    
    """
    op_type = 'MVLEM'

    def __init__(self, osi, dens, ele_nodes, m, c, thick: list=None, widths: list=None, rho: list=None, mat_concretes: list=None, mat_steels: list=None, mat_shear=None):
        r"""
        Initial method for MVLEM

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        dens: float
            Wall density
        ele_nodes: list
            A list of two element nodes
        m: int
            Number of element macro-fibers
        c: float
            Location of center of rotation from the inode, ``c`` = 0.4 (recommended)
        thick: list, optional
            A list of ``m`` macro-fiber thicknesses
        widths: list, optional
            A list of ``m`` macro-fiber widths
        rho: list, optional
            A list of m reinforcing ratios corresponding to macro-fibers; for each fiber: :math:`rho_i =
            a_{s,i}/a_{gross,i} (1 < i < m)`
        mat_concretes: list, optional
            A list of ``m`` uniaxial_material objects for concrete
        mat_steels: list, optional
            A list of ``m`` uniaxial_material objects for steel
        mat_shear: obj, optional
            Object of uniaxial_material for shear material

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> mat_conc = [o3.uniaxial_material.Concrete01(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0),
        >>>             o3.uniaxial_material.Concrete01(osi, fpc=1.0, epsc0=1.0, fpcu=1.0, eps_u=1.0)]
        >>> mat_steel = [o3.uniaxial_material.Steel02(osi, fy=1.0, e0=1.0, b=1.0, params=[15, 0.925, 0.15])]
        >>> mat_shear = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
        >>> o3.element.MVLEM(osi, dens=1.0, ele_nodes=ele_nodes, m=1, c=1.0, thick=[1.0, 1.0], widths=[1, 1], rho=[1., 1.],
        >>>                  mat_concretes=mat_conc, mat_steels=mat_steel, mat_shear=mat_shear)
        """
        self.osi = osi
        self.dens = float(dens)
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.c = float(c)
        self.thick = thick
        self.widths = widths
        self.rho = rho
        if mat_concretes is None:
            self.mat_concretes = None
        else:
            self.mat_concretes = [x.tag for x in mat_concretes]
        if mat_steels is None:
            self.mat_steels = None
        else:
            self.mat_steels = [x.tag for x in mat_steels]
        self.mat_shear = mat_shear
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, self.dens, *self.ele_node_tags, self.m, self.c]
        if getattr(self, 'thick') is not None:
            self._parameters += ['-thick', *self.thick]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'rho') is not None:
            self._parameters += ['-rho', *self.rho]
        if getattr(self, 'mat_concretes') is not None:
            self._parameters += ['-matConcrete', *self.mat_concretes]
        if getattr(self, 'mat_steels') is not None:
            self._parameters += ['-matSteel', *self.mat_steels]
        if getattr(self, 'mat_shear') is not None:
            self._parameters += ['-matShear', self.mat_shear.tag]
        self.to_process(osi)


class SFIMVLEM(ElementBase):
    """
    The SFIMVLEM Element Class
    
    The SFI_MVLEM command is used to construct a Shear-Flexure Interaction Multiple-Vertical-Line-Element Model
    (SFI-MVLEM, Kolozvari et al., 2015a, b, c), which captures interaction between axial/flexural and shear
    behavior of RC structural walls and columns under cyclic loading. The SFI_MVLEM element (Figure 1)
    incorporates 2-D RC panel behavior described by the Fixed-Strut-Angle-Model (nDMaterial FSAM;
    Ulugtekin, 2010; Orakcal et al., 2012), into a 2-D macroscopic fiber-based model (MVLEM).
    The interaction between axial and shear behavior is captured at each RC panel
    (macro-fiber) level, which further incorporates interaction between shear
    and flexural behavior at the SFI_MVLEM element level.

    
    """
    op_type = 'SFI_MVLEM'

    def __init__(self, osi, ele_nodes, m, c, thick=None, widths=None, mats=None):
        """
        Initial method for SFIMVLEM

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        m: int
            Number of element macro-fibers
        c: float
            Location of center of rotation with from the inode, ``c`` = 0.4 (recommended)
        thick: None, optional
            
        widths: None, optional
            
        mats: None, optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> mats = [o3.uniaxial_material.Elastic(osi, 1.0, 1.0), o3.uniaxial_material.Elastic(osi, 1.0, 1.0)]
        >>> mat_tags = [x.tag for x in mats]  # TODO: should pass in mats not mat tags
        >>> o3.element.SFIMVLEM(osi, ele_nodes=ele_nodes, m=1, c=1.0, thick=[1.0, 1.0], widths=[1.0, 1.0], mat_tags=mat_tags)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.m = int(m)
        self.c = float(c)
        self.thick = thick
        self.widths = widths
        self.mats = mats
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.m, self.c]
        if getattr(self, 'thick') is not None:
            self._parameters += ['-thick', *self.thick]
        if getattr(self, 'widths') is not None:
            self._parameters += ['-width', *self.widths]
        if getattr(self, 'mats') is not None:
            self._parameters += ['-mat', *self.mats]
        self.to_process(osi)



from o3seespy.base_model import OpenSeesObject


class IntegratorBase(OpenSeesObject):
    op_base_type = "integrator"


class LoadControl(IntegratorBase):
    """
    The LoadControl Integrator Class
    
    Create a OpenSees LoadControl integrator object.
    """
    op_type = 'LoadControl'

    def __init__(self, osi, incr, num_iter=1, min_incr: float=None, max_incr: float=None):
        r"""
        Initial method for LoadControl

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        incr: float
            Load factor increment :math:`\lambda`.
        num_iter: int, optional
            Number of iterations the user would like to occur in the solution algorithm. 
        min_incr: float (default=True), optional
            Min stepsize the user will allow :math:`\lambda_{min}`. 
        max_incr: float (default=True), optional
            Max stepsize the user will allow :math:`\lambda_{max}`. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.LoadControl(osi, incr=1.0, num_iter=1, min_incr=None, max_incr=None)
        """
        self.osi = osi
        self.incr = float(incr)
        self.num_iter = int(num_iter)
        if min_incr is None:
            self.min_incr = None
        else:
            self.min_incr = float(min_incr)
        if max_incr is None:
            self.max_incr = None
        else:
            self.max_incr = float(max_incr)
        self._parameters = [self.op_type, self.incr, self.num_iter]
        special_pms = ['min_incr', 'max_incr']
        packets = [False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class DisplacementControl(IntegratorBase):
    """
    The DisplacementControl Integrator Class
    
    Create a DisplacementControl integrator.  In an analysis step with Displacement Control we seek to determine the
    time step that will result in a displacement increment for a particular degree-of-freedom at a node to be a
    prescribed value.
    """
    op_type = 'DisplacementControl'

    def __init__(self, osi, node, dof, incr, num_iter=1, d_umin: float=None, d_umax: float=None):
        r"""
        Initial method for DisplacementControl

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        node: obj
            Object of node whose response controls solution
        dof: int
            Degree of freedom at the node, 1 through ndf.
        incr: float
            First displacement increment :math:`\delta u_{dof}`.
        num_iter: int, optional
            Number of iterations the user would like to occur in the solution algorithm. 
        d_umin: None (default=True), optional
            
        d_umax: None (default=True), optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> node = o3.node.Node(osi, 0.0, 0.0)
        >>> o3.integrator.DisplacementControl(osi, node, dof=1, incr=1.0, num_iter=1, d_umin=None, d_umax=None)
        """
        self.osi = osi
        self.node = node
        self.dof = int(dof)
        self.incr = float(incr)
        self.num_iter = int(num_iter)
        self.d_umin = d_umin
        self.d_umax = d_umax
        self._parameters = [self.op_type, self.node.tag, self.dof, self.incr, self.num_iter]
        special_pms = ['d_umin', 'd_umax']
        packets = [False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class ParallelDisplacementControl(IntegratorBase):
    """
    The ParallelDisplacementControl Integrator Class
    
    Create a Parallel version of DisplacementControl integrator.  In an analysis step with Displacement Control we seek
    to determine the time step that will result in a displacement increment for a particular degree-of-freedom at a node
    to be a prescribed value.
    """
    op_type = 'ParallelDisplacementControl'

    def __init__(self, osi, node, dof, incr, num_iter=1, d_umin: float=None, d_umax: float=None):
        r"""
        Initial method for ParallelDisplacementControl

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        node: obj
            Object of node whose response controls solution
        dof: int
            Degree of freedom at the node, 1 through ndf.
        incr: float
            First displacement increment :math:`\delta u_{dof}`.
        num_iter: int, optional
            Number of iterations the user would like to occur in the solution algorithm. 
        d_umin: None (default=True), optional
            
        d_umax: None (default=True), optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> node = o3.node.Node(osi, 0.0, 0.0)
        >>> o3.integrator.ParallelDisplacementControl(osi, node, dof=1, incr=1.0, num_iter=1, d_umin=None, d_umax=None)
        """
        self.osi = osi
        self.node = node
        self.dof = int(dof)
        self.incr = float(incr)
        self.num_iter = int(num_iter)
        self.d_umin = d_umin
        self.d_umax = d_umax
        self._parameters = [self.op_type, self.node.tag, self.dof, self.incr, self.num_iter]
        special_pms = ['d_umin', 'd_umax']
        packets = [False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class MinUnbalDispNorm(IntegratorBase):
    """
    The MinUnbalDispNorm Integrator Class
    
    Create a MinUnbalDispNorm integrator.
    """
    op_type = 'MinUnbalDispNorm'

    def __init__(self, osi, dlambda1, jd=1, min_lambda: float=None, max_lambda: float=None, det: float=None):
        """
        Initial method for MinUnbalDispNorm

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        dlambda1: float
            First load increment (pseudo-time step) at the first iteration in the next invocation of the analysis
            command.
        jd: int, optional
            Factor relating first load increment at subsequent time steps. 
        min_lambda: float (default=True), optional
            Min load increment. 
        max_lambda: float (default=True), optional
            Max load increment. 
        det: None (default=True), optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.MinUnbalDispNorm(osi, dlambda1=1.0, jd=1, min_lambda=None, max_lambda=None, det=None)
        """
        self.osi = osi
        self.dlambda1 = float(dlambda1)
        self.jd = int(jd)
        if min_lambda is None:
            self.min_lambda = None
        else:
            self.min_lambda = float(min_lambda)
        if max_lambda is None:
            self.max_lambda = None
        else:
            self.max_lambda = float(max_lambda)
        self.det = det
        self._parameters = [self.op_type, self.dlambda1, self.jd]
        special_pms = ['min_lambda', 'max_lambda', 'det']
        packets = [False, False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class ArcLength(IntegratorBase):
    """
    The ArcLength Integrator Class
    
    Create a ArcLength integrator. In an analysis step with ArcLength we seek to determine the time step that will
    result in our constraint equation being satisfied.
    """
    op_type = 'ArcLength'

    def __init__(self, osi, s, alpha):
        r"""
        Initial method for ArcLength

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        s: float
            The arclength.
        alpha: float
            :math:`\alpha` a scaling factor on the reference loads.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.ArcLength(osi, s=1.0, alpha=1.0)
        """
        self.osi = osi
        self.s = float(s)
        self.alpha = float(alpha)
        self._parameters = [self.op_type, self.s, self.alpha]
        self.to_process(osi)


class CentralDifference(IntegratorBase):
    r"""
    The CentralDifference Integrator Class
    
    Create a centralDifference integrator.#. The calculation of :math:`U_t + \Delta t`, is based on using the
    equilibrium equation at time t. For this reason the method is called an explicit integration method.#. If
    there is no rayleigh damping and the C matrix is 0, for a diagonal mass matrix a diagonal solver may and
    should be used.#. For stability, :math:`\frac{\Delta t}{T_n} < \frac{1}{\pi}`
    """
    op_type = 'CentralDifference'

    def __init__(self, osi):
        """
        Initial method for CentralDifference

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.CentralDifference(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class Newmark(IntegratorBase):
    """
    The Newmark Integrator Class
    
    Create a Newmark integrator.
    """
    op_type = 'Newmark'

    def __init__(self, osi, gamma, beta, form: str=None):
        r"""
        Initial method for Newmark

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        gamma: float
            :math:`\gamma` factor.
        beta: float
            :math:`\beta` factor.
        form: str, optional
            Flag to indicate which variable to be used as primary variable  * ``'d'`` -- displacement (default) *
            ``'v'`` -- velocity * ``'a'`` -- acceleration

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.Newmark(osi, gamma=1.0, beta=1.0, form=1)
        """
        self.osi = osi
        self.gamma = float(gamma)
        self.beta = float(beta)
        self.form = form
        self._parameters = [self.op_type, self.gamma, self.beta]
        if getattr(self, 'form') is not None:
            self._parameters += ['-formD', self.form]
        self.to_process(osi)


class HHT(IntegratorBase):
    """
    The HHT Integrator Class
    
    Create a Hilber-Hughes-Taylor (HHT) integrator. This is an implicit method that allows for energy dissipation and
    second order accuracy (which is not possible with the regular Newmark object). Depending on choices of input
    parameters, the method can be unconditionally stable.
    """
    op_type = 'HHT'

    def __init__(self, osi, alpha, gamma: float=None, beta: float=None):
        r"""
        Initial method for HHT

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        alpha: float
            :math:`\alpha` factor.
        gamma: float (default=True), optional
            :math:`\gamma` factor. 
        beta: float (default=True), optional
            :math:`\beta` factor. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.HHT(osi, alpha=1.0, gamma=1.0, beta=1.0)
        """
        self.osi = osi
        self.alpha = float(alpha)
        if gamma is None:
            self.gamma = None
        else:
            self.gamma = float(gamma)
        if beta is None:
            self.beta = None
        else:
            self.beta = float(beta)
        self._parameters = [self.op_type, self.alpha]
        special_pms = ['gamma', 'beta']
        packets = [False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class GeneralizedAlpha(IntegratorBase):
    r"""
    The GeneralizedAlpha Integrator Class
    
    Create a GeneralizedAlpha integrator. This is an implicit method that like the HHT method allows for high frequency
    energy dissipation and second order accuracy, i.e. :math:`\Delta t^2`. Depending on choices of input parameters, the
    method can be unconditionally stable.
    """
    op_type = 'GeneralizedAlpha'

    def __init__(self, osi, alpha_m, alpha_f, gamma: float=None, beta: float=None):
        r"""
        Initial method for GeneralizedAlpha

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        alpha_m: float
            :math:`\alpha_m` factor.
        alpha_f: float
            :math:`\alpha_f` factor.
        gamma: float (default=True), optional
            :math:`\gamma` factor. 
        beta: float (default=True), optional
            :math:`\beta` factor. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.GeneralizedAlpha(osi, alpha_m=1.0, alpha_f=1.0, gamma=1.0, beta=1.0)
        """
        self.osi = osi
        self.alpha_m = float(alpha_m)
        self.alpha_f = float(alpha_f)
        if gamma is None:
            self.gamma = None
        else:
            self.gamma = float(gamma)
        if beta is None:
            self.beta = None
        else:
            self.beta = float(beta)
        self._parameters = [self.op_type, self.alpha_m, self.alpha_f]
        special_pms = ['gamma', 'beta']
        packets = [False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)


class TRBDF2(IntegratorBase):
    """
    The TRBDF2 Integrator Class
    
    Create a TRBDF2 integrator. The TRBDF2 integrator is a composite scheme that alternates between the Trapezoidal
    scheme and a 3 point backward Euler scheme. It does this in an attempt to conserve energy and momentum, something
    Newmark does not always do.As opposed to dividing the time-step in 2 as outlined in the `Bathe2007`_, we just
    switch alternate between the 2 integration strategies,i.e. the time step in our implementation is double
    that described in the `Bathe2007`_.
    """
    op_type = 'TRBDF2'

    def __init__(self, osi):
        """
        Initial method for TRBDF2

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.TRBDF2(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class ExplicitDifference(IntegratorBase):
    r"""
    The ExplicitDifference Integrator Class
    
    Create a ExplicitDifference integrator.#. When using Rayleigh damping, the damping ratio of high vibration modes is
    overrated, and the critical time step size will be much smaller. Hence Modal damping is more suitable for this
    method.#. There should be no zero element on the diagonal of the mass matrix when using this method.#.
    Diagonal solver should be used when lumped mass matrix is used because the equations are uncoupled.#.
    For stability, :math:`\Delta t \leq \left(\sqrt{\zeta^2+1}-\zeta\right)\frac{2}{\omega}`
    """
    op_type = 'ExplicitDifference'

    def __init__(self, osi):
        """
        Initial method for ExplicitDifference

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.integrator.ExplicitDifference(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class NewmarkExplicit(IntegratorBase):
    r"""
    Pass
    """
    op_type = 'NewmarkExplicit'

    def __init__(self, osi, gamma):
        """
        Initial method for NewmarkExplicit

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        """
        self.osi = osi
        self.gamma = gamma
        self._parameters = [self.op_type, self.gamma]
        self.to_process(osi)



from o3seespy.command.nd_material.base_material import NDMaterialBase


class CycLiqCP(NDMaterialBase):
    """
    The CycLiqCP NDMaterial Class
    
    This command is used to construct a multi-dimensional material object that that follows the constitutive behavior of
    a cyclic elastoplasticity model for large post- liquefaction deformation.CycLiqCP material is a cyclic elastoplasticity
    model for large post-liquefaction deformation, and is implemented using a cutting plane algorithm. The model is
    capable of reproducing small to large deformation in the pre- to post-liquefaction regime. The elastic moduli
    of the model are pressure dependent. The plasticity in the model is developed within the framework of
    bounding surface plasticity, with special consideration to the formulation of reversible and
    irreversible dilatancy.The model does not take into consideration of the state of sand, and
    requires different parameters for sand under different densities and confining pressures.
    The surfaces (i.e. failure and maximum pre-stress) are considered as circles in the pi
    plane.The model has been validated against VELACS centrifuge model tests and has used
    on numerous simulations of liquefaction related problems.When this material is
    employed in regular solid elements (e.g., FourNodeQuad, Brick), it simulates
    drained soil response. When solid-fluid coupled elements (u-p elements and
    SSP u-p elements) are used, the model is able to simulate undrained and partially drained behavior of soil.
    """
    op_type = 'CycLiqCP'

    def __init__(self, osi, g0, kappa, h, mfc, dre1, mdc, dre2, rdr, alpha, dir, ein, rho):
        """
        Initial method for CycLiqCP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        g0: float
            A constant related to elastic shear modulus
        kappa: float
            Bulk modulus
        h: float
            Model parameter for plastic modulus
        mfc: float
            Stress ratio at failure in triaxial compression
        dre1: float
            Coefficient for reversible dilatancy generation
        mdc: float
            Stress ratio at which the reversible dilatancy sign changes
        dre2: float
            Coefficient for reversible dilatancy release
        rdr: float
            Reference shear strain length
        alpha: float
            Parameter controlling the decrease rate of irreversible dilatancy
        dir: float
            Coefficient for irreversible dilatancy potential
        ein: float
            Initial void ratio
        rho: float
            Saturated mass density

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.CycLiqCP(osi, g0=1.0, kappa=1.0, h=1.0, mfc=1.0, dre1=1.0, mdc=1.0, dre2=1.0, rdr=1.0, alpha=1.0, dir=1.0, ein=1.0, rho=1.0)
        """
        self.osi = osi
        self.g0 = float(g0)
        self.kappa = float(kappa)
        self.h = float(h)
        self.mfc = float(mfc)
        self.dre1 = float(dre1)
        self.mdc = float(mdc)
        self.dre2 = float(dre2)
        self.rdr = float(rdr)
        self.alpha = float(alpha)
        self.dir = float(dir)
        self.ein = float(ein)
        self.rho = float(rho)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.g0, self.kappa, self.h, self.mfc, self.dre1, self.mdc, self.dre2, self.rdr, self.alpha, self.dir, self.ein, self.rho]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_update_material_stage(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)


class CycLiqCPSP(NDMaterialBase):
    """
    The CycLiqCPSP NDMaterial Class
    
    This command is used to construct a multi-dimensional material object that that follows the constitutive behavior of
    a cyclic elastoplasticity model for large post- liquefaction deformation.CycLiqCPSP material is a constitutive model
    for sand with special considerations for cyclic behaviour and accumulation of large post-liquefaction shear
    deformation, and is implemented using a cutting plane algorithm. The model: (1) achieves the simulation of
    post-liquefaction shear deformation based on its physics, allowing the unified description of pre- and
    post-liquefaction behavior of sand; (2) directly links the cyclic mobility of sand with reversible
    and irreversible dilatancy, enabling the unified description of monotonic and cyclic loading; (3)
    introduces critical state soil mechanics concepts to achieve unified modelling of sand under
    different states.The critical, maximum stress ratio and reversible dilatancy surfaces
    follow a rounded triangle in the pi plane similar to the Matsuoka-Nakai
    criterion.When this material is employed in regular solid elements
    (e.g., FourNodeQuad, Brick), it simulates drained soil response.
    When solid-fluid coupled elements (u-p elements and SSP u-p
    elements) are used, the model is able to simulate undrained and partially drained behavior of soil.
    """
    op_type = 'CycLiqCPSP'

    def __init__(self, osi, g0, kappa, h, big_m, dre1, dre2, rdr, alpha, dir, lambdac, ksi, e0, np, nd, ein, rho):
        """
        Initial method for CycLiqCPSP

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        g0: float
            A constant related to elastic shear modulus
        kappa: float
            Bulk modulus
        h: float
            Model parameter for plastic modulus
        big_m: float
            Critical state stress ratio
        dre1: float
            Coefficient for reversible dilatancy generation
        dre2: float
            Coefficient for reversible dilatancy release
        rdr: float
            Reference shear strain length
        alpha: float
            Parameter controlling the decrease rate of irreversible dilatancy
        dir: float
            Coefficient for irreversible dilatancy potential
        lambdac: float
            Critical state constant
        ksi: float
            Critical state constant
        e0: float
            Void ratio at pc=0
        np: float
            Material constant for peak mobilized stress ratio
        nd: float
            Material constant for reversible dilatancy generation stress ratio
        ein: float
            Initial void ratio
        rho: float
            Saturated mass density

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.CycLiqCPSP(osi, g0=1.0, kappa=1.0, h=1.0, big_m=1.0, dre1=1.0, dre2=1.0, rdr=1.0, alpha=1.0, dir=1.0, lambdac=1.0, ksi=1.0, e0=1.0, np=1.0, nd=1.0, ein=1.0, rho=1.0)
        """
        self.osi = osi
        self.g0 = float(g0)
        self.kappa = float(kappa)
        self.h = float(h)
        self.big_m = float(big_m)
        self.dre1 = float(dre1)
        self.dre2 = float(dre2)
        self.rdr = float(rdr)
        self.alpha = float(alpha)
        self.dir = float(dir)
        self.lambdac = float(lambdac)
        self.ksi = float(ksi)
        self.e0 = float(e0)
        self.np = float(np)
        self.nd = float(nd)
        self.ein = float(ein)
        self.rho = float(rho)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.g0, self.kappa, self.h, self.big_m, self.dre1, self.dre2, self.rdr, self.alpha, self.dir, self.lambdac, self.ksi, self.e0, self.np, self.nd, self.ein, self.rho]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_update_material_stage(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)



from o3seespy.opensees_instance import OpenSeesInstance
from o3seespy import exceptions
from o3seespy import extensions
from o3seespy import cc
from o3seespy import cc as static  # deprecated
from o3seespy.command import node, algorithm, rayleigh, test, uniaxial_material, element, nd_material
from o3seespy.command.common import *
from o3seespy.command import section, beam_integration, constraints, numberer, system, region, friction_model
from o3seespy.command import integrator, analysis, recorder, pattern, time_series, geom_transf, patch, layer
from .command import mesh, senscmds
import o3seespy.tools
from o3seespy.command import test_check  # deprecated
from o3seespy.__about__ import __version__
from o3seespy import results
from . import mptools, mp

try:
    from custom_openseespy import custom_opensees as opy
except ModuleNotFoundError:
    import openseespy.opensees as opy


import o3seespy as o3  # for testing only


def test_catenary_cable():
    osi = o3.OpenSeesInstance(ndm=2)
    i_node = o3.node.Node(osi, 0.0, 0.0)
    j_node = o3.node.Node(osi, 0.0, 1.0)
    o3.element.CatenaryCable(osi, i_node=i_node, j_node=j_node, weight=1.0, big_e=1.0, big_a=1.0, l0=1.0, alpha=1.0, temperature_change=1.0, rho=1.0, error_tol=1.0, nsubsteps=1, mass_type=1)




import o3seespy as o3  # for testing only
import pytest


@pytest.mark.skip()
def test_pfem_element_bubble():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.PFEMElementBubble(osi, ele_nodes, rho=1.0, mu=1.0, b1=1.0, b2=1.0, b3=1.0, thickness=1.0, kappa=1.0)


@pytest.mark.skip()
def test_pfem_element_compressible():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
    o3.element.PFEMElementCompressible(osi, ele_nodes, rho=1.0, mu=1.0, b1=1.0, b2=1.0, thickness=1.0, kappa=1.0)




from o3seespy.command.element.base_element import ElementBase


class ZeroLength(ElementBase):
    """
        The ZeroLength Element Class

        This command is used to construct a zeroLength element object, which is defined by two nodes at the same location.
        The nodes are connected by multiple UniaxialMaterial objects to represent the force-deformation relationship for the
        element.
        """
    op_type = 'zeroLength'

    def __init__(self, osi, node_i, node_j, mat=None, mat_x=None, mat_y=None, mat_z=None,
                 mat_rot_x=None, mat_rot_y=None, mat_rot_z=None, r_flag=0):
        self.node_i = node_i
        self.node_j = node_j
        self.mat = mat
        self.mats = [mat_x, mat_y, mat_z, mat_rot_x, mat_rot_y, mat_rot_z]
        self.r_flag = r_flag
        osi.n_ele += 1
        self._tag = osi.n_ele

        self._parameters = [self.op_type, self._tag, *[self.node_i.tag, self.node_j.tag]]
        mats = []
        dirs = []
        for i, mat in enumerate(self.mats):
            if mat is not None:
                mats.append(mat.tag)
                dirs.append(i + 1)
        if len(mats) == 1:
            mats = mats[0]
            dirs = dirs[0]
        elif len(mats) == 0:
            if self.mat is None:
                raise ValueError("Must set either mat or a mat direction")
            mats = self.mat.tag
            dirs = 1
        self._parameters.append("-mat")
        self._parameters.append(mats)
        self._parameters.append("-dir")
        self._parameters.append(dirs)
        self._parameters.append("-doRayleigh")
        self._parameters.append(self.r_flag)
        self.to_process(osi)



from o3seespy.base_model import OpenSeesObject


class ConstraintsBase(OpenSeesObject):
    op_base_type = "constraints"


class Plain(ConstraintsBase):
    """
    The Plain Constraints Class
    
    This command is used to construct a Plain constraint handler. A plain constraint handler can only enforce
    homogeneous single point constraints (fix command) and multi-point constraints constructed where the
    constraint matrix is equal to the identity (equalDOF command). The following is the command to
    construct a plain constraint handler:.. note::As mentioned, this constraint handler can only
    enforce homogeneous single point constraints (fix command) and multi-pont constraints where
    the constraint matrix is equal to the identity (equalDOF command).
    """
    op_type = 'Plain'

    def __init__(self, osi):
        """
        Initial method for Plain

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.constraints.Plain(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)


class Lagrange(ConstraintsBase):
    """
    The Lagrange Constraints Class
    
    This command is used to construct a LagrangeMultiplier constraint handler, which enforces the constraints by
    introducing Lagrange multiplies to the system of equation. The following is the command to construct a plain
    constraint handler:
    """
    op_type = 'Lagrange'

    def __init__(self, osi, alpha_s=1.0, alpha_m=1.0):
        r"""
        Initial method for Lagrange

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        alpha_s: float, optional
            :math:`\alpha_s` factor on single points.
        alpha_m: float, optional
            :math:`\alpha_m` factor on multi-points.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.constraints.Lagrange(osi, alpha_m=1.0)
        """
        self.osi = osi
        self.alpha_s = float(alpha_s)
        self.alpha_m = float(alpha_m)
        self._parameters = [self.op_type, self.alpha_s, self.alpha_m]
        self.to_process(osi)


class Penalty(ConstraintsBase):
    """
    The Penalty Constraints Class
    
    This command is used to construct a Penalty constraint handler, which enforces the constraints using the penalty
    method. The following is the command to construct a penalty constraint handler:
    """
    op_type = 'Penalty'

    def __init__(self, osi, alpha_s=1.0, alpha_m=1.0):
        r"""
        Initial method for Penalty

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        alpha_s: float, optional
            :math:`\alpha_s` factor on single points.
        alpha_m: float, optional
            :math:`\alpha_m` factor on multi-points.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.constraints.Penalty(osi, alpha_m=1.0)
        """
        self.osi = osi
        self.alpha_s = float(alpha_s)
        self.alpha_m = float(alpha_m)
        self._parameters = [self.op_type, self.alpha_s, self.alpha_m]
        self.to_process(osi)


class Transformation(ConstraintsBase):
    """
    The Transformation Constraints Class
    
    This command is used to construct a transformation constraint handler, which enforces the constraints using the
    transformation method. The following is the command to construct a transformation constraint handler.. note::* The
    single-point constraints when using the transformation method are done directly. The matrix equation is not
    manipulated to enforce them, rather the trial displacements are set directly at the nodes at the start of
    each analysis step.* Great care must be taken when multiple constraints are being enforced as the
    transformation method does not follow constraints:#. If a node is fixed, constrain it with the
    fix command and not equalDOF or other type of constraint.#. If multiple nodes are
    constrained, make sure that the retained node is not constrained in any other
    constraint.And remember if a node is constrained to multiple nodes in your model it probably means you have messed up.
    """
    op_type = 'Transformation'

    def __init__(self, osi):
        """
        Initial method for Transformation

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.constraints.Transformation(osi)
        """
        self.osi = osi
        self._parameters = [self.op_type]
        self.to_process(osi)



from o3seespy.base_model import OpenSeesObject


class PatchBase(OpenSeesObject):
    op_base_type = "patch"


class Quad(PatchBase):
    """
    The Quad Patch Class
    
    The patch command is used to generate a number of fibers over a cross-sectional area. Currently there are three
    types of cross-section that fibers can be generated: quadrilateral, rectangular and circular.
    """
    op_type = 'quad'

    def __init__(self, osi, mat, num_subdiv_ij, num_subdiv_jk, crds_i, crds_j, crds_k, crds_l):
        """
        Initial method for Quad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat: obj
            Material object associated with this fiber (uniaxial_material object for a fibersection and ndmaterial
            object for use in an ndfibersection).
        num_subdiv_ij: int
            Number of subdivisions (fibers) in the ij direction.
        num_subdiv_jk: int
            Number of subdivisions (fibers) in the jk direction.
        crds_i: list
            Y & z-coordinates of vertex i (local coordinate system)
        crds_j: list
            Y & z-coordinates of vertex j (local coordinate system)
        crds_k: list
            Y & z-coordinates of vertex k (local coordinate system)
        crds_l: list
            Y & z-coordinates of vertex l (local coordinate system)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> crds_i = [1.0, 1.0]
        >>> crds_j = [1.0, 1.0]
        >>> crds_k = [1.0, 1.0]
        >>> crds_l = [1.0, 1.0]
        >>> conc_conf = o3.uniaxial_material.Concrete01(osi, fpc=-5.0, epsc0=-0.005, fpcu=-3.5, eps_u=-0.02)
        >>> o3.patch.Quad(osi, conc_conf, num_subdiv_ij=1, num_subdiv_jk=1, crds_i=crds_i, crds_j=crds_j, crds_k=crds_k, crds_l=crds_l)
        """
        self.osi = osi
        self.mat = mat
        self.num_subdiv_ij = int(num_subdiv_ij)
        self.num_subdiv_jk = int(num_subdiv_jk)
        self.crds_i = crds_i
        self.crds_j = crds_j
        self.crds_k = crds_k
        self.crds_l = crds_l
        self._parameters = [self.op_type, self.mat.tag, self.num_subdiv_ij, self.num_subdiv_jk, *self.crds_i, *self.crds_j, *self.crds_k, *self.crds_l]
        self.to_process(osi)

class Rect(PatchBase):
    """
    The Rect Patch Class
    
    This is the command to generate a rectangular patch. The geometry of the patch is defined by coordinates of
    vertices: I and J. The first vertex, I, is the bottom-left point and the second vertex, J, is the top-right
    point, having as a reference the local y-z plane.
    """
    op_type = 'rect'

    def __init__(self, osi, mat, num_subdiv_y, num_subdiv_z, crds_i, crds_j):
        """
        Initial method for Rect

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat: obj
            Material object associated with this fiber (uniaxial_material object for a fibersection and ndmaterial
            object for use in an ndfibersection).
        num_subdiv_y: int
            Number of subdivisions (fibers) in local y direction.
        num_subdiv_z: int
            Number of subdivisions (fibers) in local z direction.
        crds_i: list
            Y & z-coordinates of vertex i (local coordinate system)
        crds_j: list
            Y & z-coordinates of vertex j (local coordinate system)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> crds_i = [1.0, 1.0]
        >>> crds_j = [1.0, 1.0]
        >>> conc_conf = o3.uniaxial_material.Concrete01(osi, fpc=-5.0, epsc0=-0.005, fpcu=-3.5, eps_u=-0.02)
        >>> o3.patch.Rect(osi, conc_conf, num_subdiv_y=1, num_subdiv_z=1, crds_i=crds_i, crds_j=crds_j)
        """
        self.osi = osi
        self.mat = mat
        self.num_subdiv_y = int(num_subdiv_y)
        self.num_subdiv_z = int(num_subdiv_z)
        self.crds_i = crds_i
        self.crds_j = crds_j
        self._parameters = [self.op_type, self.mat.tag, self.num_subdiv_y, self.num_subdiv_z, *self.crds_i, *self.crds_j]
        self.to_process(osi)

class Circ(PatchBase):
    """
    The Circ Patch Class
    
    This is the command to generate a circular shaped patch
    """
    op_type = 'circ'

    def __init__(self, osi, mat, num_subdiv_circ, num_subdiv_rad, center, rad, ang):
        """
        Initial method for Circ

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat: obj
            Material object associated with this fiber (uniaxial_material object for a fibersection and ndmaterial
            object for use in an ndfibersection).
        num_subdiv_circ: int
            Number of subdivisions (fibers) in the circumferential direction (number of wedges)
        num_subdiv_rad: int
            Number of subdivisions (fibers) in the radial direction (number of rings)
        center: list
            Y & z-coordinates of the center of the circle
        rad: list
            Internal & external radius
        ang: list
            Starting & ending-coordinates angles (degrees)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> center = [1.0, 1.0]
        >>> rad = [1.0, 1.0]
        >>> ang = [1.0, 1.0]
        >>> conc_conf = o3.uniaxial_material.Concrete01(osi, fpc=-5.0, epsc0=-0.005, fpcu=-3.5, eps_u=-0.02)
        >>> o3.patch.Circ(osi, conc_conf, num_subdiv_circ=1, num_subdiv_rad=1, center=center, rad=rad, ang=ang)
        """
        self.osi = osi
        self.mat = mat
        self.num_subdiv_circ = int(num_subdiv_circ)
        self.num_subdiv_rad = int(num_subdiv_rad)
        self.center = center
        self.rad = rad
        self.ang = ang
        self._parameters = [self.op_type, self.mat.tag, self.num_subdiv_circ, self.num_subdiv_rad, *self.center, *self.rad, *self.ang]
        self.to_process(osi)



from o3seespy.base_model import OpenSeesObject


class BeamIntegrationBase(OpenSeesObject):
    op_base_type = "beamIntegration"


class Lobatto(BeamIntegrationBase):
    """
    The Lobatto BeamIntegration Class
    
    Create a Gauss-Lobatto beamIntegration object.Gauss-Lobatto integration is the most common approach for evaluating
    the response of`forceBeamColumn-Element` (`Neuenhofer and Filippou 1997`_) because it places an integration point at
    each end of the element, where bending moments are largest in the absence of interior element loads.
    """
    op_type = 'Lobatto'

    def __init__(self, osi, sec, big_n):
        """
        Initial method for Lobatto

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec: obj
            A previous-defined section object.
        big_n: int
            Number of integration points along the element.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.Lobatto(osi, sec=sec, big_n=1)
        """
        self.osi = osi
        self.sec = sec
        self.big_n = int(big_n)
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec.tag, self.big_n]
        self.to_process(osi)


class Legendre(BeamIntegrationBase):
    """
    The Legendre BeamIntegration Class
    
    Create a Gauss-Legendre beamIntegration object.Gauss-Legendre integration is more accurate than Gauss-Lobatto;
    however, it is not commonin force-based elements because there are no integration points at the element ends.of
    each integration point are tabulated in references on numerical analysis.The force deformation response at
    each integration point is defined by the section.The order of accuracy for Gauss-Legendre integration is
    2N-1.Arguments and examples see `Lobatto-BeamIntegration`.
    """
    op_type = 'Legendre'

    def __init__(self, osi, sec, big_n):
        """
        Initial method for Legendre

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec: obj
            
        big_n: None
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.Legendre(osi, sec=sec, big_n=1)
        """
        self.osi = osi
        self.sec = sec
        self.big_n = big_n
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec.tag, self.big_n]
        self.to_process(osi)


class NewtonCotes(BeamIntegrationBase):
    """
    The NewtonCotes BeamIntegration Class
    
    Create a Newton-Cotes beamIntegration object.Newton-Cotes places integration points uniformly along the element,
    including a point ateach end of the element.spaced integration points are tabulated in references on numerical
    analysis. The force deformationresponse at each integration point is defined by the section.The order of
    accuracy for Gauss-Radau integration is N-1.Arguments and examples see `Lobatto-BeamIntegration`.
    """
    op_type = 'NewtonCotes'

    def __init__(self, osi, sec, big_n):
        """
        Initial method for NewtonCotes

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec: obj
            
        big_n: None
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.NewtonCotes(osi, sec=sec, big_n=1)
        """
        self.osi = osi
        self.sec = sec
        self.big_n = big_n
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec.tag, self.big_n]
        self.to_process(osi)


class Radau(BeamIntegrationBase):
    """
    The Radau BeamIntegration Class
    
    Create a Gauss-Radau beamIntegration object.Gauss-Radau integration is not common in force-based elements because it
    places an integration point at only one end of the element; however, it forms the basis for optimal plastichinge
    integration methods.numerical analysis. The force-deformation response at each integration point is definedby
    the section. The order of accuracy for Gauss-Radau integration is 2N-2.Arguments and examples see
    `Lobatto-BeamIntegration`.
    """
    op_type = 'Radau'

    def __init__(self, osi, sec, big_n):
        """
        Initial method for Radau

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec: obj
            
        big_n: None
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.Radau(osi, sec=sec, big_n=1)
        """
        self.osi = osi
        self.sec = sec
        self.big_n = big_n
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec.tag, self.big_n]
        self.to_process(osi)


class Trapezoidal(BeamIntegrationBase):
    """
    The Trapezoidal BeamIntegration Class
    
    Create a Trapezoidal beamIntegration object.Arguments and examples see `Lobatto-BeamIntegration`.
    """
    op_type = 'Trapezoidal'

    def __init__(self, osi, sec, big_n):
        """
        Initial method for Trapezoidal

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec: obj
            
        big_n: None
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.Trapezoidal(osi, sec=sec, big_n=1)
        """
        self.osi = osi
        self.sec = sec
        self.big_n = big_n
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec.tag, self.big_n]
        self.to_process(osi)


class CompositeSimpson(BeamIntegrationBase):
    """
    The CompositeSimpson BeamIntegration Class
    
    Create a CompositeSimpson beamIntegration object.Arguments and examples see `Lobatto-BeamIntegration`.
    """
    op_type = 'CompositeSimpson'

    def __init__(self, osi, sec, big_n):
        """
        Initial method for CompositeSimpson

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec: obj
            
        big_n: None
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.CompositeSimpson(osi, sec=sec, big_n=1)
        """
        self.osi = osi
        self.sec = sec
        self.big_n = big_n
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec.tag, self.big_n]
        self.to_process(osi)


class UserDefined(BeamIntegrationBase):
    """
    The UserDefined BeamIntegration Class
    
    Create a UserDefined beamIntegration object.This option allows user-specified locations and weights of the
    integration points.
    """
    op_type = 'UserDefined'

    def __init__(self, osi, big_n, secs, locs, wts):
        """
        Initial method for UserDefined

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        big_n: int
            Number of integration points along the element.
        secs: list
            A list previous-defined section objects.
        locs: list
            Locations of integration points along the element.
        wts: list
            Weights of integration points.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> secs = [o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0),
        >>>          o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0)]
        >>> o3.beam_integration.UserDefined(osi, big_n=2, secs=secs, locs=[0.2, 0.9], wts=[0.5, 0.5])
        """
        self.osi = osi
        self.big_n = int(big_n)
        self.secs = [x.tag for x in secs]
        self.locs = locs
        self.wts = wts
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.big_n, *self.secs, *self.locs, *self.wts]
        self.to_process(osi)


class FixedLocation(BeamIntegrationBase):
    r"""
    The FixedLocation BeamIntegration Class
    
    Create a FixedLocation beamIntegration object.This option allows user-specified locations of the integration points.
    The associated integrationweights are computed by the method of undetermined coefficients (Vandermondesystem)

    .. math::

      \sum^N_{i=1}x_i^{j-1}w_i = \int_0^1x^{j-1}dx = \frac{1}{j},\qquad (j=1,...,N)


    Note that `NewtonCotes-BeamIntegration` integration is recovered when the integration point locations are equally
    spaced.
    """
    op_type = 'FixedLocation'

    def __init__(self, osi, big_n, secs, locs):
        """
        Initial method for FixedLocation

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        big_n: int
            Number of integration points along the element.
        secs: list
            A list previous-defined section objects.
        locs: list
            Locations of integration points along the element.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> secs = [o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0),
        >>>         o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0)]
        >>> o3.beam_integration.FixedLocation(osi, big_n=2, secs=secs, locs=[0.2, 0.9])
        """
        self.osi = osi
        self.big_n = int(big_n)
        self.secs = [x.tag for x in secs]
        self.locs = locs
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.big_n, *self.secs, *self.locs]
        self.to_process(osi)


class LowOrder(BeamIntegrationBase):
    r"""
    The LowOrder BeamIntegration Class
    
    Create a LowOrder beamIntegration object.This option is a generalization of the `FixedLocation-BeamIntegration` and
    `UserDefined-BeamIntegration` integration approaches and is useful for moving load analysis (`Kidarsa, Scott and
    Higgins 2008`_). The locations of the integration points are user defined,while a selected number of weights
    are specified and the remaining weights arecomputed by the method of undetermined coefficients.

    .. math::

      \sum_{i=1}^{N_f}x_{fi}^{j-1}w_{fi}=\frac{1}{j}-\sum_{i=1}^{N_c}x_{ci}^{j-1}w_{ci}
    """
    op_type = 'LowOrder'

    def __init__(self, osi, big_n, secs, locs, wts):
        """
        Initial method for LowOrder

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        big_n: int
            Number of integration points along the element.
        secs: list
            A list previous-defined section objects.
        locs: list
            Locations of integration points along the element.
        wts: list
            Weights of integration points.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> secs = [o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0),
        >>>         o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0)]
        >>> o3.beam_integration.LowOrder(osi, big_n=2, secs=secs, locs=[0.2, 0.9], wts=[0.5, 0.5])
        """
        self.osi = osi
        self.big_n = int(big_n)
        self.secs = [x.tag for x in secs]
        self.locs = locs
        self.wts = wts
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.big_n, *self.secs, *self.locs, *self.wts]
        self.to_process(osi)


class MidDistance(BeamIntegrationBase):
    r"""
    The MidDistance BeamIntegration Class
    
    Create a MidDistance beamIntegration object.This option allows user-specified locations of the integration points.
    The associated integration weights are determined from the midpoints between adjacent integration point
    locations.:math:`w_i=(x_{i+1}-x_{i-1})/2` for :math:`i=2...N-1`, :math:`w_1=(x_1+x_2)/2`, and
    :math:`w_N=1-(x_{N-1}+x_N)/2`.
    """
    op_type = 'MidDistance'

    def __init__(self, osi, big_n, secs, locs):
        """
        Initial method for MidDistance

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        big_n: int
            Number of integration points along the element.
        secs: list
            A list previous-defined section objects.
        locs: list
            Locations of integration points along the element.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> secs = [o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0),
        >>>         o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0)]
        >>> o3.beam_integration.MidDistance(osi, big_n=2, secs=secs, locs=[0.2, 0.9])
        """
        self.osi = osi
        self.big_n = int(big_n)
        self.secs = [x.tag for x in secs]
        self.locs = locs
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.big_n, *self.secs, *self.locs]
        self.to_process(osi)


class UserHinge(BeamIntegrationBase):
    """
    The UserHinge BeamIntegration Class
    
    Create a UserHinge beamIntegration object.
    """
    op_type = 'UserHinge'

    def __init__(self, osi, sec_e, np_l, secs_ls, locs_l, wts_l, np_r, secs_rs, locs_r, wts_r):
        """
        Initial method for UserHinge

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec_e: obj
            A previous-defined section objects for non-hinge area.
        np_l: int
            Number of integration points along the left hinge.
        secs_ls: list
            A list of previous-defined section objects for left hinge area.
        locs_l: list
            A list of locations of integration points for left hinge area.
        wts_l: list
            A list of weights of integration points for left hinge area.
        np_r: int
            Number of integration points along the right hinge.
        secs_rs: list
            A list of previous-defined section objects for right hinge area.
        locs_r: list
            A list of locations of integration points for right hinge area.
        wts_r: list
            A list of weights of integration points for right hinge area.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec_e = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> secs_l = [o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)]
        >>> secs_r = [o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)]
        >>> o3.beam_integration.UserHinge(osi, sec_e=sec_e, np_l=1, secs_ls=secs_l, locs_l=[1], wts_l=[1], np_r=1,
        >>>                               secs_rs=secs_r, locs_r=[1], wts_r=[1])
        """
        self.osi = osi
        self.sec_e = sec_e
        self.np_l = int(np_l)
        self.secs_ls = [x.tag for x in secs_ls]
        self.locs_l = locs_l
        self.wts_l = wts_l
        self.np_r = int(np_r)
        self.secs_rs = [x.tag for x in secs_rs]
        self.locs_r = locs_r
        self.wts_r = wts_r
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec_e.tag, self.np_l, *self.secs_ls, *self.locs_l, *self.wts_l, self.np_r, *self.secs_rs, *self.locs_r, *self.wts_r]
        self.to_process(osi)


class HingeMidpoint(BeamIntegrationBase):
    """
    The HingeMidpoint BeamIntegration Class
    
    Create a HingeMidpoint beamIntegration object.Midpoint integration over each hinge region is the most accurate
    one-point integration rule;however, it does not place integration points at the element ends and there is a small
    integrationerror for linear curvature distributions along the element.
    """
    op_type = 'HingeMidpoint'

    def __init__(self, osi, sec_i, lp_i, sec_j, lp_j, sec_e):
        """
        Initial method for HingeMidpoint

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec_i: obj
            A previous-defined section object for hinge at i.
        lp_i: float
            The plastic hinge length at i.
        sec_j: obj
            A previous-defined section object for hinge at j.
        lp_j: float
            The plastic hinge length at j.
        sec_e: obj
            A previous-defined section object for the element interior.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec_i = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> sec_j = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> sec_e = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.HingeMidpoint(osi, sec_i=sec_i, lp_i=1.0, sec_j=sec_j, lp_j=1.0, sec_e=sec_e)
        """
        self.osi = osi
        self.sec_i = sec_i
        self.lp_i = float(lp_i)
        self.sec_j = sec_j
        self.lp_j = float(lp_j)
        self.sec_e = sec_e
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec_i.tag, self.lp_i, self.sec_j.tag, self.lp_j, self.sec_e.tag]
        self.to_process(osi)


class HingeRadau(BeamIntegrationBase):
    """
    The HingeRadau BeamIntegration Class
    
    Create a HingeRadau beamIntegration object.Modified two-point Gauss-Radau integration over each hinge region places
    an integration point atthe element ends and at 8/3 the hinge length inside the element. This approach representslinear
    curvature distributions exactly and the characteristic length for softening plastic hinges is equal to the assumed
    palstic hinge length.Arguments and examples see `HingeMidPoint-BeamIntegration`.
    """
    op_type = 'HingeRadau'

    def __init__(self, osi, sec_i, lp_i, sec_j, lp_j, sec_e):
        """
        Initial method for HingeRadau

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec_i: obj
            
        lp_i: None
            
        sec_j: obj
            
        lp_j: None
            
        sec_e: obj
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec_i = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> sec_j = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> sec_e = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.HingeRadau(osi, sec_i=sec_i, lp_i=1.0, sec_j=sec_j, lp_j=1.0, sec_e=sec_e)
        """
        self.osi = osi
        self.sec_i = sec_i
        self.lp_i = lp_i
        self.sec_j = sec_j
        self.lp_j = lp_j
        self.sec_e = sec_e
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec_i.tag, self.lp_i, self.sec_j.tag, self.lp_j, self.sec_e.tag]
        self.to_process(osi)


class HingeRadauTwo(BeamIntegrationBase):
    """
    The HingeRadauTwo BeamIntegration Class
    
    Create a HingeRadauTwo beamIntegration object.Two-point Gauss-Radau integration over each hinge region places an
    integrationpoint at the element ends and at 2/3 the hinge length inside the element. This approachrepresents linear
    curvature distributions exactly; however, the characteristic length for softeningplastic hinges is not equal to
    the assumed plastic hinge length (equals 1/4 of the plastic hinge length).Arguments and examples see
    `HingeMidPoint-BeamIntegration`.
    """
    op_type = 'HingeRadauTwo'

    def __init__(self, osi, sec_i, lp_i, sec_j, lp_j, sec_e):
        """
        Initial method for HingeRadauTwo

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec_i: obj
            
        lp_i: None
            
        sec_j: obj
            
        lp_j: None
            
        sec_e: obj
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec_i = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> sec_j = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> sec_e = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.HingeRadauTwo(osi, sec_i=sec_i, lp_i=1.0, sec_j=sec_j, lp_j=1.0, sec_e=sec_e)
        """
        self.osi = osi
        self.sec_i = sec_i
        self.lp_i = lp_i
        self.sec_j = sec_j
        self.lp_j = lp_j
        self.sec_e = sec_e
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec_i.tag, self.lp_i, self.sec_j.tag, self.lp_j, self.sec_e.tag]
        self.to_process(osi)


class HingeEndpoint(BeamIntegrationBase):
    """
    The HingeEndpoint BeamIntegration Class
    
    Create a HingeEndpoint beamIntegration object.Endpoint integration over each hinge region moves the integration
    points to the element ends;however, there is a large integration error for linear curvature distributions along
    the element.
    """
    op_type = 'HingeEndpoint'

    def __init__(self, osi, sec_i, lp_i, sec_j, lp_j, sec_e):
        """
        Initial method for HingeEndpoint

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        sec_i: obj
            A previous-defined section object for hinge at i.
        lp_i: float
            The plastic hinge length at i.
        sec_j: obj
            A previous-defined section object for hinge at j.
        lp_j: float
            The plastic hinge length at j.
        sec_e: obj
            A previous-defined section object for the element interior.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> sec_i = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> sec_j = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> sec_e = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.beam_integration.HingeEndpoint(osi, sec_i=sec_i, lp_i=1.0, sec_j=sec_j, lp_j=1.0, sec_e=sec_e)
        """
        self.osi = osi
        self.sec_i = sec_i
        self.lp_i = float(lp_i)
        self.sec_j = sec_j
        self.lp_j = float(lp_j)
        self.sec_e = sec_e
        osi.n_integ += 1
        self._tag = osi.n_integ
        self._parameters = [self.op_type, self._tag, self.sec_i.tag, self.lp_i, self.sec_j.tag, self.lp_j, self.sec_e.tag]
        self.to_process(osi)
